import uuid
from itertools import takewhile
from os import scandir
from pathlib import Path
import collections
import colorlog
import copy
import logging
import sys
from typing import Union, Iterable

import numpy as np

log = logging.getLogger('__name__')

LOG_FORMAT_STR = u'%(asctime)s %(levelname)-8s %(filename)s:%(lineno)-4d %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
LOG_COLORS = {
    'DEBUG': 'green',
    'INFO': 'cyan',
    'WARNING': 'bold_yellow',
    'ERROR': 'bold_red',
    'CRITICAL': 'bold_purple'}


class Bunch(dict):
    """A subclass of dictionary with an additional dot syntax."""

    def __init__(self, *args, **kwargs):
        super(Bunch, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self, deep=False):
        """Return a new Bunch instance which is a copy of the current Bunch instance.

        Parameters
        ----------
        deep : bool
            If True perform a deep copy (see notes). By default a shallow copy is returned.

        Returns
        -------
        Bunch
            A new copy of the Bunch.

        Notes
        -----
        - A shallow copy constructs a new Bunch object and then (to the extent possible) inserts
        references into it to the objects found in the original.
        - A deep copy constructs a new Bunch and then, recursively, inserts copies into it of the
         objects found in the original.
        """
        return copy.deepcopy(self) if deep else Bunch(super(Bunch, self).copy())

    def save(self, npz_file, compress=False):
        """
        Saves a npz file containing the arrays of the bunch.

        :param npz_file: output file
        :param compress: bool (False) use compression
        :return: None
        """
        if compress:
            np.savez_compressed(npz_file, **self)
        else:
            np.savez(npz_file, **self)

    @staticmethod
    def load(npz_file):
        """
        Loads a npz file containing the arrays of the bunch.

        :param npz_file: output file
        :return: Bunch
        """
        if not Path(npz_file).exists():
            raise FileNotFoundError(f"{npz_file}")
        return Bunch(np.load(npz_file))


def _iflatten(x):
    result = []
    for el in x:
        if isinstance(el, collections.abc.Iterable) and not (
                isinstance(el, str) or isinstance(el, dict)):
            result.extend(_iflatten(el))
        else:
            result.append(el)
    return result


def _gflatten(x):
    def iselement(e):
        return not (isinstance(e, collections.abc.Iterable) and not (
            isinstance(el, str) or isinstance(el, dict)))
    for el in x:
        if iselement(el):
            yield el
        else:
            yield from _gflatten(el)


def flatten(x, generator=False):
    """
    Flatten a nested Iterable excluding strings and dicts.

    Converts nested Iterable into flat list. Will not iterate through strings or
    dicts.

    :return: Flattened list or generator object.
    :rtype: list or generator
    """
    return _gflatten(x) if generator else _iflatten(x)


def range_str(values: iter) -> str:
    """
    Given a list of integers, returns a terse string expressing the unique values.

    Example:
        indices = [0, 1, 2, 3, 4, 7, 8, 11, 15, 20]
        range_str(indices)
        >> '0-4, 7-8, 11, 15 & 20'
    :param values: An iterable of ints
    :return: A string of unique value ranges
    """
    trial_str = ''
    values = list(set(values))
    for i in range(len(values)):
        if i == 0:
            trial_str += str(values[i])
        elif values[i] - (values[i - 1]) == 1:
            if i == len(values) - 1 or values[i + 1] - values[i] > 1:
                trial_str += f'-{values[i]}'
        else:
            trial_str += f', {values[i]}'
    # Replace final comma with an ampersand
    k = trial_str.rfind(',')
    if k > -1:
        trial_str = f'{trial_str[:k]} &{trial_str[k + 1:]}'
    return trial_str


def setup_logger(name='ibl', level=logging.NOTSET, file=None, no_color=False):
    """Set up a log for IBL packages.

    Uses date time, calling function and distinct colours for levels.
    Sets the name if not set already and add a stream handler.
    If the stream handler already exists, does not duplicate.
    The naming/level allows not to interfere with third-party libraries when setting level.

    Parameters
    ----------
    name : str
        Log name, should be set to the root package name for consistent logging throughout the app.
    level : str, int
        The logging level (defaults to NOTSET, which inherits the parent log level)
    file : bool, str, pathlib.Path
        If True, a file handler is added with the default file location, otherwise a log file path
        may be passed.
    no_color : bool
        If true the colour log is deactivated.  May be useful when directing the std out to a file.

    Returns
    -------
    logging.Logger, logging.RootLogger
        The configured log.
    """
    log = logging.getLogger() if not name else logging.getLogger(name)
    log.setLevel(level)
    fkwargs = {'no_color': True} if no_color else {'log_colors': LOG_COLORS}
    # check existence of stream handlers before adding another
    if not any(map(lambda x: x.name == f'{name}_auto', log.handlers)):
        # need to remove any previous default Stream handler configured on stderr
        # to not duplicate output
        for h in log.handlers:
            if isinstance(h, logging.StreamHandler) and h.stream.name == '<stderr>' and h.level == 0 and h.name is None:
                log.removeHandler(h)
        stream_handler = logging.StreamHandler(stream=sys.stdout)
        stream_handler.setFormatter(
            colorlog.ColoredFormatter('%(log_color)s' + LOG_FORMAT_STR,
                                      LOG_DATE_FORMAT, **fkwargs))
        stream_handler.name = f'{name}_auto'
        log.addHandler(stream_handler)
    # add the file handler if requested, but check for duplicates
    if not any(map(lambda x: x.name == f'{name}_file', log.handlers)):
        if file is True:
            log_to_file(log=name)
        elif file:
            log_to_file(filename=file, log=name)
    return log


def log_to_file(log='ibl', filename=None):
    """
    Save log information to a given filename in '.ibl_logs' folder (in home directory).

    Parameters
    ----------
    log : str, logging.Logger
        The log (name or object) to add file handler to.
    filename : str, Pathlib.Path
        The name of the log file to save to.

    Returns
    -------
    logging.Logger
        The log with the file handler attached.
    """
    if isinstance(log, str):
        log = logging.getLogger(log)
    if filename is None:
        filename = Path.home().joinpath('.ibl_logs', log.name)
    elif not Path(filename).is_absolute():
        filename = Path.home().joinpath('.ibl_logs', filename)
    filename.parent.mkdir(exist_ok=True)
    file_handler = logging.FileHandler(filename, encoding='utf-8')
    file_format = logging.Formatter(LOG_FORMAT_STR, LOG_DATE_FORMAT)
    file_handler.setFormatter(file_format)
    file_handler.name = f'{log.name}_file'
    log.addHandler(file_handler)
    log.info(f'File log initiated {file_handler.name}')
    return log


def rrmdir(folder: Path, levels: int = 0):
    """
    Recursively remove a folder and its parents up to a defined level - if they are empty.

    Parameters
    ----------
    folder : pathlib.Path
        The path to a folder at which to start the recursion.
    levels : int
        Recursion level, i.e. the number of parents to delete, relative to `folder`.
        Defaults to 0 - which has the same effect as `pathlib.Path.rmdir` except that it won't
        raise an OSError if the directory is not empty.

    Returns
    -------
    list of pathlib.Path
        A list of folders that were recursively removed.

    Raises
    ------
    FileNotFoundError
        If `folder` does not exist.
    PermissionError
        Insufficient privileges or folder in use by another process.
    NotADirectoryError
        The folder provided is most likely a file.
    """
    try:  # a sorted list of absolute nested folder paths
        to_remove = (folder, *folder.parents[:levels])  # py >= 3.9
    except TypeError:  # py <= 3.8 compatible
        to_remove = (folder, *[folder.parents[n] for n in range(levels)])
    # filter list to those that are empty; if statement always true as rmdir returns None
    return [f for f in takewhile(lambda f: not any(f.iterdir()), to_remove) if not f.rmdir()]


def dir_size(directory: Union[str, Path], follow_symlinks: bool = False) -> int:
    """
    Calculate the total size of a directory including all its subdirectories and files.

    Parameters
    ----------
    directory : str | Path
        The path to the directory for which the size needs to be calculated.
    follow_symlinks : bool, optional
        Whether to follow symbolic links when calculating the size. Default is False.

    Returns
    -------
    int
        The total size of the directory in bytes.
    """
    total_bytes = 0
    with scandir(directory) as it:
        for entry in it:
            if entry.is_symlink() and not follow_symlinks:
                continue
            elif entry.is_dir():
                total_bytes += dir_size(entry.path, follow_symlinks)
            elif entry.is_file():
                total_bytes += entry.stat().st_size
    return total_bytes


def get_mac() -> str:
    """
    Fetch the machine's unique MAC address formatted according to IEEE 802 specifications.

    Returns
    -------
    str
        The MAC address of the device formatted in six groups of two
        hexadecimal digits separated by hyphens in transmission order
        (e.g., 'BA-DB-AD-C0-FF-EE').
    """
    return uuid.getnode().to_bytes(6, 'big').hex('-').upper()


def ensure_list(value, exclude_type=(str, dict)):
    """Ensure input is a list.

    Wraps `value` in a list if not already an iterator or if it is a member of specific
    iterable classes.

    To allow users the option of passing a single value or multiple values, this function
    will wrap the former in a list and by default will consider str and dict instances as
    a single value. This function is useful because it treats tuples, lists, sets, and
    generators all as 'lists', but not dictionaries and strings.

    Parameters
    ----------
    value : any
        Input to ensure list.
    exclude_type : tuple, optional
        A list of iterable classes to wrap in a list.

    Returns
    -------
    Iterable
        Either `value` if iterable and not in `exclude_type` list, or `value` wrapped in a list.
    """
    return [value] if isinstance(value, exclude_type) or not isinstance(value, Iterable) else value



"""
Examples
--------

# Connect to remote server rig, send initialization message and wait for response
>>> server = await EchoProtocol.server('udp://192.168.0.4', name='main')
>>> await server.init('2022-01-01_1_subject')  # Send init message and await confirmation of receipt
>>> response = await server.on_event('INIT')  # Await response

# Send initialization message and wait max 10 seconds for response
>>> try:
...     response = await asyncio.wait_for(server.on_event('INIT'), 10.)
... except asyncio.TimeoutError:
...     server.close()
"""
import sys
import json
import urllib.parse
import urllib.request
import asyncio
import argparse
import socket
import logging
from types import MappingProxyType
from collections import UserDict
from functools import partial

import colorlog

from iblutil.io.net import base
from iblutil import util


def _setup_log(name, level=logging.DEBUG):
    """A colour log with the log name in the format string"""
    log = logging.getLogger(name)
    log.setLevel(level)

    log.handlers = []
    fmt_str = '%(log_color)s%(asctime)s [%(name)s] %(levelname)-8s %(filename)s:%(lineno)-4d %(message)s'
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(
        colorlog.ColoredFormatter(fmt_str, util.LOG_DATE_FORMAT, log_colors=util.LOG_COLORS))
    stream_handler.name = f'{name}_auto'
    stream_handler.setLevel(level)
    log.addHandler(stream_handler)
    return log


def _address2tuple(address) -> (str, int):
    """Convert URI to (host, port) tuple.

    Convert URI to form used by transport layer.

    Parameters
    ----------
    address : str
        A URI from which to extract the port and hostname.

    Returns
    -------
    str
        The hostname.
    int
        The port.
    """
    server_uri = base.validate_uri(address, default_port=base.LISTEN_PORT)
    parsed_uri = urllib.parse.urlparse(server_uri)
    return parsed_uri.hostname, parsed_uri.port


class EchoProtocol(base.Communicator):
    """An echo server implementing TCP/IP and UDP.

    This should be instantiated using either EchoProtocol.server or EchoProtocol.client.
    In the client role, the remote address is specified; in the server role, the local address is
    specified.

    Attributes
    ----------
    Server : asyncio.Server
        A network server instance if using TCP/IP.
    role : {'client', 'server'}
        The communicator role.  A server may communicate with multiple clients. The server URI
        specifies its local address.  A client only communicates with a single host, specified by
        the server URI.
    default_echo_timeout : float
        The default maximum time in seconds to await a message echo.
    _last_sent : dict[(str, int), (bytes, asyncio.Future)]
        A map of addresses holding the last sent bytes str and the future being waited on.  In
        client mode there should only be one entry - the server URI.
    """

    Server = None
    _role = None
    default_echo_timeout = 1.

    def __init__(self, server_uri, role, name=None, logger=None):
        super().__init__(server_uri, name=name, logger=logger)
        self._transport = None
        self._socket = None
        self.role = role
        # For validating echo'd response
        self._last_sent = {}
        # Transport specific futures
        loop = asyncio.get_running_loop()
        self.on_connection_lost = loop.create_future()
        self.on_error_received = loop.create_future()
        self.on_eof_received = loop.create_future()

    @property
    def role(self) -> str:
        """{'client', 'server'}: The remote computer's role"""
        return self._role

    @role.setter
    def role(self, value: str):
        """Set remote computer role.

        Ensures the role is only set once.

        Parameters
        ----------
        value : {'client', 'server'}
            The role to set.

        Raises
        ------
        AttributeError
            The role has already been set and cannot be changed.
        ValueError
            The role must be one of {'client', 'server'}.
        """
        if self._role is not None:
            raise AttributeError('can\'t set attribute')
        if (value or '').strip().lower() not in ('server', 'client'):
            raise ValueError('role must be either "server" or "client"')
        self._role = value.strip().lower()

    @property
    def is_connected(self) -> bool:
        """bool: True if transport layer set and open."""
        return self._transport and not self._transport.is_closing()

    def awaiting_response(self, addr=None) -> bool:
        """bool: True if awaiting confirmation of receipt from remote."""
        if addr:
            last_sent = self._last_sent.get(addr, False)
            return last_sent and not last_sent[1].done()
        else:
            return self._last_sent and any(not x[1].done() for x in self._last_sent.values())

    async def cleanup(self, data=None, addr=None):
        """Cleanup experiment.

        Send a cleanup message to the remote host.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().cleanup(data)
        await self.confirmed_send(message, addr=addr)

    async def start(self, exp_ref, data=None, addr=None):
        """Start an experiment.

        Send a stop message to the remote host.

        Parameters
        ----------
        exp_ref : str
            A experiment reference string in the form yyyy-mm-dd_n_subject.
        data : any
            Optional extra data to send to the remote host.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().start(exp_ref, data)
        await self.confirmed_send(message, addr=addr)

    async def stop(self, data=None, immediately=False, addr=None):
        """End an experiment.

        Send a stop message to the remote host.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        immediately : bool
            If True, an EXPINTERRUPT signal is used.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().stop(data, immediately=immediately)
        await self.confirmed_send(message, addr=addr)

    async def status(self, status, addr=None):
        """Communicate experiment status.

        Send a status message to the remote host.

        Parameters
        ----------
        status : iblutil.net.base.ExpStatus
            An experiment status enumeration.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().status(status)
        await self.confirmed_send(message, addr=addr)

    async def info(self, status, data=None, addr=None):
        """Communicate experiment information.

        Send experiment status and details to the remote host.

        Parameters
        ----------
        status : iblutil.net.base.ExpStatus
            An experiment status enumeration.
        data : any
            Optional extra data to send to the remote host.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().info(status, data)
        await self.confirmed_send(message, addr=addr)

    async def init(self, data=None, addr=None):
        """Initialize an experiment.

        Send an initialization message to the remote host.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        addr : (str, int)
            The remote host address and port. Only required in server role.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        """
        message = super().init(data)
        await self.confirmed_send(message, addr=addr)

    async def alyx(self, alyx=None, addr=None):
        """
        Send/request Alyx token to/from remote host.

        Parameters
        ----------
        alyx : one.webclient.AlyxClient
            An instance of Alyx to extract and send token from.
        addr : (str, int)
        The remote host address and port. Only required in server role.

        Returns
        -------
        (str, dict) or None
            (If alyx arg was None or not authenticated) the received Alyx token in the form (base_url, {user: token}).
        """
        message = super().alyx(alyx)
        if all(message):  # send instance to remote host
            await self.confirmed_send(message, addr=addr)
        else:  # request instance from remote host
            loop = asyncio.get_running_loop()
            fut = loop.create_future()
            self.assign_callback('ALYX', fut)
            await self.confirmed_send(message, addr=addr)
            token, _ = await fut
            return token

    @staticmethod
    def encode(data):
        """Serialize data for transmission"""
        if isinstance(data, (bytes, bytearray)):
            return data
        if not isinstance(data, str):
            data = json.dumps(data)
        return data.encode()

    def send(self, data, addr=None):
        """Send data to clients.

        Serialize data and pass to transport layer.
        """
        super().send(data, addr=addr)
        if self.protocol == 'udp':
            addr = addr or (self.hostname, self.port)
            self.logger.debug(f'Send "{data}" to udp://{addr[0]}:{addr[1]}')
            self._transport.sendto(data, addr)
        else:
            addr = addr or self._socket.getpeername()
            if addr != self._socket.getpeername():
                self.logger.warning(f'Message not sent: unexpected address {addr}')
                return
            self.logger.debug(f'Send "{data}" to {self.protocol}://{addr[0]}:{addr[1]}')
            self._transport.write(self.encode(data))
            # self._transport.write_eof()

    async def confirmed_send(self, data, addr=None, timeout=None):
        """
        Send a message to the client and await echo.

        NB: Methods such as start, stop, init, cleanup and alyx should be used instead of calling
        this directly.

        Parameters
        ----------
        data : any
            The data to serialize and send to remote host.
        addr : (str, int)
            The remote host address and port. Only required in server role.
        timeout : float, optional
            The time in seconds to wait for an echo before raising an exception.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
        RuntimeError
            The response from the client did not match the original message.
        ValueError
            Timeout must be non-zero number.
            Unexpected remote address: in client mode the address must match server_uri.
        TypeError
            In server mode a remote address must be provided.
        """
        if self.role == 'server':
            if not addr:
                raise TypeError('confirmed_send missing 1 required argument: \'addr\'')
        elif addr and addr != (self.hostname, self.port):
            raise ValueError('Unexpected remote address')
        addr = addr or (self.hostname, self.port)
        if not (timeout := timeout or self.default_echo_timeout) > 0:
            raise ValueError('Timeout must be non-zero number')
        loop = asyncio.get_running_loop()
        echo_future = loop.create_future()
        # echo_future.add_done_callback(lambda _: self._last_sent.pop(addr))  # delete below instead (no difference)
        self._last_sent[addr] = (self.encode(data), echo_future)
        self.send(self._last_sent[addr][0], addr=addr)
        # Sockets can no longer be blocking, so we'll wait ourselves.
        try:
            await asyncio.wait_for(echo_future, timeout=timeout)
        except asyncio.TimeoutError:
            self.close()
            raise TimeoutError(f'Failed to receive client response in time ({self.name}: {self.server_uri})')
        except RuntimeError:
            self.close()
            raise RuntimeError('Unexpected response from server')
        del self._last_sent[addr]

    def close(self):
        """
        Close the connection, de-register callbacks and cancel outstanding futures.

        The EchoProtocol.on_connection_lost future is resolved at this time, all others are
        cancelled.  NB: Closing the socket should be handled by transport base class later on.
        """
        # Close transport
        if self._transport:
            self._transport.close()

        super().close()  # Deregister callbacks, cancel event futures
        echo_futures = map(lambda x: x[1], self._last_sent.values())
        for fut in filter(lambda f: not f.done(), echo_futures):
            fut.cancel('Close called on communicator')
        if self.on_error_received and not self.on_error_received.done():
            self.on_error_received.cancel('Close called on communicator')
        if self.on_eof_received and not self.on_eof_received.done():
            self.on_eof_received.cancel('Close called on communicator')
        if self.on_connection_lost and not self.on_connection_lost.done():
            self.on_connection_lost.set_result('Close called on communicator')

    # The following methods are inherited from asyncio.DatagramProtocol and called by the event loop

    def connection_made(self, transport):
        """Called by event loop"""
        self._transport = transport
        self._socket = transport.get_extra_info('socket')

        # Validate
        if self._socket.type is socket.SOCK_DGRAM:
            if self.protocol != 'udp':
                raise RuntimeError('Unsupported transport layer for UDP')
        elif self._socket.type is socket.SOCK_STREAM:
            if self.protocol not in ('ws', 'wss', 'tcp'):
                raise RuntimeError('Unsupported transport layer for TCP/IP')
        else:
            raise RuntimeError(f'Unsupported transport layer with socket type "{self._socket.type.name}"')
        self.logger.debug(f'Connected with socket {self._socket}')

    def _receive(self, data, addr):
        """
        Process data received from remote host.  This is called by different lower level methods
        depending on the transport layer.  This method handles the message echo logic, while the
        base class method handles message callbacks exclusively.

        If awaiting an echo, the timeout is cancelled and the message is checked against the one
        sent.  Otherwise, the message is immediately echo'd and the super class method is called to
        notify any listeners.

        This method should not be called by the user.

        Parameters
        ----------
        data : bytes
            The serialized data received by the transport layer.
        addr : (str, int)
            The source address as (hostname, port)

        Notes
        -----
        - Callbacks apply only to new messages, not echo responses.  EchoProtocol.confirmed_send
        should be awaited, however it is possible to assign a callback for receipt of an echo with
        EchoProtocol._last_sent[addr][1].add_done_callback.
        - Currently this only checks the received message, not its origin.
        """
        host, port = addr[:2]
        msg = data.decode()
        # If we're still awaiting echo from this address, process here. NB: sometimes the echo_future is already set
        # but not yet removed from last sent, so we also check future not successfully finished.
        if (last_sent := self._last_sent.get(addr)) and not base.is_success(last_sent[1]):
            expected, echo_future = last_sent
            # If echo doesn't match, raise exception
            if data != expected:
                self.logger.error('Expected %s from %s, got %s', expected, self.name, data)
                echo_future.set_exception(RuntimeError)
            elif not echo_future.cancelled():  # Notify callbacks of receipt
                self.logger.info('Confirmation received')
                echo_future.set_result(True)
        else:
            self.logger.info('Received %r from %s://%s:%i', msg, self.protocol, host, port)
            # Update from remote
            if msg[1] != '0':  # do not echo 0 code messages; these are low-level errors
                self.logger.debug('Echo %r to %s://%s:%i', msg, self.protocol, host, port)
                self.send(data, addr)  # Echo
            super()._receive(data, addr)  # Process callbacks

    def datagram_received(self, data, addr):
        """Called by UDP transport layer"""
        host, port = addr[:2]
        if self.role == 'client' and host != self.hostname:
            self.logger.warning(
                f'Ignoring UDP packet from unexpected host ({host}:{port}) with message "{data}"')
        else:
            self._receive(data, addr)

    def data_received(self, data):
        """Called by TCP/IP transport layer"""
        addr = self._transport.get_extra_info('peername')[:2]
        self._receive(data, addr)

    def error_received(self, exc):
        self.logger.error('Error received: %s', exc)
        self.on_error_received.set_result(exc)

    def eof_received(self):
        self.logger.debug('EOF received')
        self.on_eof_received.set_result(True)

    def connection_lost(self, exc):
        self.logger.info('Connection closed')
        if getattr(self, 'on_con_lost', False):
            self.on_connection_lost.set_result(exc)

    # Factory methods for instantiating a server or client

    @staticmethod
    async def server(server_uri, name=None, log=None, **kwargs) -> 'EchoProtocol':
        """
        Create a remote server instance.

        Parameters
        ----------
        server_uri : str, ipaddress.IPv4Address, ipaddress.IPv6Address
            The address of the remote computer, may be an IP or hostname with or without a port.
            To use TCP/IP instead of the default UDP, add a 'ws://' scheme to the URI.
        name : str
            An optional, arbitrary label.
        kwargs
            Optional parameters to pass to create_datagram_endpoint for UDP or create_server for
            TCP/IP.

        Returns
        -------
        EchoProtocol
            A Communicator instance.
        """
        # Validate server URI
        server_uri = base.validate_uri(server_uri)

        # Get a reference to the event loop
        loop = asyncio.get_running_loop()

        # Logging
        log = log or _setup_log(name or server_uri)

        # One protocol instance will be created to serve all client requests.
        if server_uri.startswith('udp'):
            Protocol = partial(EchoProtocol, server_uri, 'server', name=name, logger=log)
            _, protocol = await loop.create_datagram_endpoint(Protocol, local_addr=_address2tuple(server_uri), **kwargs)
        else:
            protocol = EchoProtocol(server_uri, 'server', name=name, logger=log)
            protocol.Server = await loop.create_server(lambda: protocol, *_address2tuple(server_uri), **kwargs)

        protocol.logger.info(f'Listening on {protocol.server_uri}')
        return protocol

    @staticmethod
    async def client(server_uri, name=None, log=None, **kwargs) -> 'EchoProtocol':
        """
        Create a remote client instance.

        Parameters
        ----------
        server_uri : str
            The address of the remote computer, may be an IP or hostname with or without a port.
            To use TCP/IP instead of the default UDP, add a 'ws://' scheme to the URI.
        name : str
            An optional, arbitrary label.
        kwargs
            Optional parameters to pass to create_datagram_endpoint for UDP or create_server for
            TCP/IP.

        Returns
        -------
        EchoProtocol
            A Communicator instance.
        """
        # Validate server URI
        server_uri = base.validate_uri(server_uri)

        # Get a reference to the event loop
        loop = asyncio.get_running_loop()

        # Logging
        log = log or _setup_log(name or server_uri)

        Protocol = partial(EchoProtocol, server_uri, 'client', name=name, logger=log)
        if server_uri.startswith('udp'):
            _, protocol = await loop.create_datagram_endpoint(Protocol, remote_addr=_address2tuple(server_uri), **kwargs)
        else:
            _, protocol = await loop.create_connection(Protocol, *_address2tuple(server_uri), **kwargs)

        return protocol


class Services(base.Service, UserDict):
    """Handler for multiple remote rig services."""
    __slots__ = ('timeout', 'server')

    def __init__(self, remote_rigs, timeout=10.):
        """Handler for multiple remote rig services.

        Parameters
        ----------
        remote_rigs : list(iblutil.io.net.base.Service)
            A list of remote rig service objects.
        timeout : float
            How long to wait for response from client(s).
        """
        # Store rig communicators by name
        super().__init__()
        if not all(isinstance(x, base.Service) for x in remote_rigs):
            raise TypeError(f'Remote services must be of type {type(base.Service)}')
        self.data = MappingProxyType({rig.name: rig for rig in remote_rigs})  # Ensure immutable
        self.timeout = timeout

    @property
    def is_connected(self) -> bool:
        """bool: All services are connected."""
        return any(self.values()) and all(map(lambda rig: rig.is_connected, self.values()))

    def assign_callback(self, event, callback, return_service=False):
        """
        Assign a callback to all services for a given event.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            An event to listen for.
        callback : function, async.io.Future
            A callable or future to notify when the event occurs.
        return_service : bool
            When True an instance of the Communicator is additionally passed to the callback.
        """
        def _callback(service, data, addr):
            callback(data, addr, service)

        for service in self.values():
            if return_service:
                cb = partial(_callback, service)
                # keep track of original callback id
                cb.id = getattr(callback, 'id', None) or hash(callback)
                service.assign_callback(event, cb)
            else:
                service.assign_callback(event, callback)

    def clear_callbacks(self, event, callback=None):
        """
        Clear all callbacks for a given event.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            The event to clear listeners from.
        callback : function, asyncio.Future
            A specific callback or future to remove.
        """
        removed = {}
        for rig in self.values():
            removed[rig.name] = rig.clear_callbacks(event, callback=callback)
        return removed

    async def await_all(self, event):
        """
        Wait for all services to report a given event.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            The event to wait on.

        Returns
        -------
        dict
            A map of rig name and the data that was received.
        """
        responses = {}

        async def _return_data(rig, response) -> None:
            """Return map of rig name and data so we know origin of data"""
            data, _ = await response
            responses[rig.name] = data
            return response

        event = base.ExpMessage.validate(event)
        tasks = set()
        for rig in self.values():
            task = asyncio.create_task(_return_data(rig, rig.on_event(event)))
            tasks.add(task)

        if self.timeout:  # py3.11 with asyncio.timeout context manager
            _, pending = await asyncio.wait(
                tasks, timeout=self.timeout, return_when=asyncio.ALL_COMPLETED
            )
            if any(pending):
                failed = set(self.keys()).difference(responses.keys())
                raise asyncio.TimeoutError(
                    f'The following services failed to respond in time: {failed}')
        else:
            await asyncio.gather(*tasks)

        return responses

    def close(self):
        """Close all communication."""
        for rig in self.values():
            rig.close()

    async def init(self, data=None, concurrent=True):
        """Initialize an experiment.

        Send an initialization signal to the remote services and await the responses.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        concurrent : bool
            If false, wait for response from each service before communicating with the next.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage['EXPINIT']
        return await self._signal(event, 'init', data=data, concurrent=concurrent)

    async def cleanup(self, data=None, concurrent=True):
        """Cleanup an experiment.

        Send a cleanup signal to the remote services and await responses.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        concurrent : bool
            If false, wait for response from each service before communicating with the next.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage.EXPCLEANUP
        return await self._signal(event, 'cleanup', data=data, reverse=True, concurrent=concurrent)

    async def start(self, exp_ref, data=None, concurrent=True):
        """Start an experiment.

        Send a start signal to the remote services and await responses.

        Parameters
        ----------
        exp_ref : str
            An experiment reference string in the form yyyy-mm-dd_n_subject.
        data : any
            Optional extra data to send to the remote host.
        concurrent : bool
            If false, wait for response from each service before communicating with the next.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage.EXPSTART
        return await self._signal(event, 'start', exp_ref, data=data, concurrent=concurrent)

    async def stop(self, data=None, immediately=False, **kwargs):
        """End an experiment.

        Send a stop signal to the remote services and await responses.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote host.
        immediately : bool
            If true, send an EXPINTERRUPT signal.
        concurrent : bool
            If false, wait for response from each service before communicating with the next.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage.EXPEND
        return await self._signal(event, 'stop', data=data, immediately=immediately, reverse=True, **kwargs)

    async def status(self, status, **kwargs):
        """Communicate experiment status.

        Send a status message to the remote services and await responses.

        Parameters
        ----------
        status : iblutil.net.base.ExpStatus
            An experiment status enumeration.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage.EXPSTATUS
        return await self._signal(event, 'status', status, **kwargs)

    async def info(self, status, data=None, **kwargs):
        """Report experiment information.

        Send a status message and other details to the remote services and await responses.

        Parameters
        ----------
        status : iblutil.net.base.ExpStatus
            An experiment status enumeration.
        data : any
            Optional extra data to send to the remote host.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        Raises
        ------
        TimeoutError
            Remote host failed to echo the message within the timeout period.
            Remote host failed to respond within response period.
        """
        event = base.ExpMessage.EXPINFO
        return await self._signal(event, 'info', status, data=data, **kwargs)

    async def _signal(self, event, method, *args, concurrent=True, reverse=False, **kwargs):
        """Send an event signal to the remote services and await responses.

        Parameters
        ----------
        event : iblutil.io.net.base.ExpMessage
            The event to signal to services.
        method : str
            The name of the method to call for each service.
        args
            Positional arguments to pass to method.
        concurrent : bool
            If true, all services are signaled concurrently.
        reverse : bool
            If true, iterate over services in reverse order.
        kwargs
            Keyword arguments to pass to method.

        Returns
        -------
        dict of str
            A dictionary of service names and the response data received.

        TODO Could initialize response dict created in await_all and allow it to be returned
         this would allow one to peek at responses before all are in
        """
        event = base.ExpMessage.validate(event, allow_bitwise=False)
        if concurrent:
            # Register event callbacks before sending messages otherwise we may receive a response before callback is
            # created.
            all_responses = asyncio.create_task(self.await_all(event), name='service responses')
            for service in (reversed(list(self.values())) if reverse else self.values()):
                f = getattr(service, method or event.name.lower())
                await f(*args, **kwargs)
            responses = await all_responses
        else:
            responses = dict.fromkeys(self.keys())
            for name, service in (reversed(list(self.items())) if reverse else self.items()):
                f = getattr(service, method or event.name.lower())
                await f(*args, **kwargs)
                if self.timeout:
                    data, _ = await asyncio.wait_for(service.on_event(event), self.timeout)
                else:
                    data, _ = await service.on_event(event)
                responses[service.name] = data
        return responses

    async def alyx(self, alyx):
        """
        Send Alyx token to remote services.

        Parameters
        ----------
        alyx : one.webclient.AlyxClient
            An instance of Alyx to extract and send token from.
        """
        for service in self.values():
            await service.alyx(alyx)


async def main(role, server_uri, name=None, **kwargs):
    """An example of an entry point for creating an individual communicator."""
    if role == 'server':
        print('Starting server')
        com = await EchoProtocol.server(server_uri, name=name, **kwargs)
        try:
            if com.server_uri.startswith('udp'):
                await asyncio.sleep(60 * 60)  # Serve for 1 hour.
            else:
                await com.Server.serve_forever()
        finally:
            com.close()
    elif role == 'client':
        print('Starting client')
        com = await EchoProtocol.client(server_uri, name=name)
        try:
            # Here you would send a message
            await com.start('2022-01-01_1_subject')
        finally:
            com.close()

    else:
        raise ValueError(f'Unknown role "{role}"')


if __name__ == '__main__':
    """
    Examples
    --------
    # Run server
    python udp.py server

    # Run server locally for debugging
    python udp.py server -H localhost

    # Run client for debugging
    python udp.py client
    """
    # Parse parameters
    parser = argparse.ArgumentParser(description='UDP Experiment Communicator.')
    parser.add_argument('role', choices=('server', 'client'),
                        help='communicator role i.e. server or client')
    parser.add_argument('--host', '-H', help='the host address', default=base.hostname2ip())
    parser.add_argument('--verbose', '-v', action='count', default=0)
    args = parser.parse_args()  # returns data from the options specified

    asyncio.run(main(args.role, args.host))



from pathlib import Path, PurePath
from datetime import datetime
import collections
import sys
import os
import json
import subprocess
import logging
import time
import socket
import asyncio
from math import inf


def as_dict(par):
    if not par or isinstance(par, dict):
        return par
    else:
        return dict(par._asdict())


def from_dict(par_dict):
    if not par_dict:
        return None
    par = collections.namedtuple('Params', par_dict.keys())

    class IBLParams(par):
        __slots__ = ()

        def set(self, field, value):
            d = as_dict(self)
            d[field] = value
            return from_dict(d)

        def as_dict(self):
            return as_dict(self)

    return IBLParams(**par_dict)


def getfile(str_params):
    """
    Returns full path of the param file per system convention:
     linux/mac: ~/.str_params, Windows: APPDATA folder

    :param str_params: string that identifies parm file
    :return: string of full path
    """
    # strips already existing dot if any
    parts = ['.' + p if not p.startswith('.') else p for p in Path(str_params).parts]
    if sys.platform == 'win32' or sys.platform == 'cygwin':
        pfile = str(PurePath(os.environ['APPDATA'], *parts))
    else:
        pfile = str(Path.home().joinpath(*parts))
    return pfile


def set_hidden(path, hide: bool) -> Path:
    """
    Set a given file or folder path to be hidden.  On macOS and Windows a specific flag is set,
    while on other systems the file or folder is simply renamed to start with a dot.  On macOS the
    folder may only be hidden in Explorer.

    Parameters
    ----------
    path : str, pathlib.Path
        The path of the file or folder to (un)hide.
    hide : bool
        If True the path is set to hidden, otherwise it is unhidden.

    Returns
    -------
    pathlib.Path
        The path of the file or folder, which may have been renamed.
    """
    path = Path(path)
    assert path.exists()
    if sys.platform == 'win32' or sys.platform == 'cygwin':
        flag = ('+' if hide else '-') + 'H'
        subprocess.run(['attrib', flag, str(path)]).check_returncode()
    elif sys.platform == 'darwin':
        flag = ('' if hide else 'no') + 'hidden'
        subprocess.run(['chflags', flag, str(path)]).check_returncode()
    elif hide and not path.name.startswith('.'):
        path = path.rename(path.parent.joinpath('.' + path.name))
    elif not hide and path.name.startswith('.'):
        path = path.rename(path.parent.joinpath(path.name[1:]))
    return path


def read(str_params, default=None):
    """
    Reads in and parse Json parameter file into dictionary.  If the parameter file doesn't
    exist and no defaults are provided, a FileNotFound error is raised, otherwise any extra
    default parameters will be written into the file.

    Examples:
        # Load parameters, raise error if file not found
        par = read('globus/admin')

        # Load with defaults
        par = read('globus/admin', {'local_endpoint': None, 'remote_endpoint': None})

        # Return empty dict if file not found (i.e. touch new param file)
        par = read('new_pars', {})

    :param str_params: path to text json file
    :param default: default values for missing parameters
    :return: named tuple containing parameters
    """
    pfile = getfile(str_params)
    par_dict = as_dict(default) or {}
    if Path(pfile).exists():
        with open(pfile) as fil:
            file_pars = json.loads(fil.read())
        par_dict.update(file_pars)
    elif default is None:  # No defaults provided
        raise FileNotFoundError(f'Parameter file {pfile} not found')

    if not Path(pfile).exists() or par_dict.keys() > file_pars.keys():
        # write the new parameter file with the extra param
        write(str_params, par_dict)
    return from_dict(par_dict)


def write(str_params, par):
    """
    Write a parameter file in Json format

    :param str_params: path to text json file
    :param par: dictionary containing parameters values
    :return: None
    """
    pfile = Path(getfile(str_params))
    if not pfile.parent.exists():
        pfile.parent.mkdir()
    dpar = as_dict(par)
    for k in dpar:
        if isinstance(dpar[k], Path):
            dpar[k] = str(dpar[k])
    with open(pfile, 'w') as fil:
        json.dump(as_dict(par), fil, sort_keys=False, indent=4)


class FileLock:
    def __init__(self, filename, log=None, timeout=10, timeout_action='delete'):
        """
        A context manager to ensure a file is not written to.

        This context manager checks whether a lock file already exists, indicating that the
        filename is currently being written to by another process, and waits until it is free
        before entering.  If the lock file is not removed within the timeout period, it is either
        forcebly removed (assumes other process hanging or killed), or raises an exception.

        Before entering, a new lock file is created, containing the hostname, datetime and pid,
        then subsequenctly removed upon exit.

        Parameters
        ----------
        filename : pathlib.Path, str
            A filepath to 'lock'.
        log : logging.Logger
            A logger instance to use.
        timeout : float
            How long to wait before either raising an exception or deleting the previous lock file.
        timeout_action : {'delete', 'raise'} str
            Action to take if previous lock file remains throughout timeout period. Either delete
            the old lock file or raise an exception.

        Examples
        --------
        Ensure a file is not being written to by another process before writing

        >>> with FileLock(filename, timeout_action='delete'):
        >>>     with open(filename, 'w') as fp:
        >>>         fp.write(r'{"foo": "bar"}')

        Asychronous implementation example with raise behaviour

        >>> try:
        >>>     async with FileLock(filename, timeout_action='raise'):
        >>>         with open(filename, 'w') as fp:
        >>>             fp.write(r'{"foo": "bar"}')
        >>> except asyncio.TimeoutError:
        >>>     print(f'failed to write to {filename}')
        """
        self.filename = Path(filename)
        self._logger = log or __name__
        if not isinstance(log, logging.Logger):
            self._logger = logging.getLogger(self._logger)

        self.timeout = timeout
        self.timeout_action = timeout_action
        if self.timeout_action not in ('delete', 'raise'):
            raise ValueError(f'Invalid timeout action: {self.timeout_action}')
        self._async_poll_freq = 0.2  # how long to sleep between lock file checks in async mode

    @property
    def lockfile(self):
        """pathlib.Path: the lock filepath."""
        return self.filename.with_suffix('.lock')

    async def _lock_check_async(self):
        while self.lockfile.exists():
            assert self._async_poll_freq > 0
            await asyncio.sleep(self._async_poll_freq)

    def __enter__(self):
        # if a lock file exists retries n times to see if it exists
        attempts = 0
        n_attempts = 5 if self.timeout else inf
        timeout = (self.timeout / n_attempts) if self.timeout else self._poll_freq

        while self.lockfile.exists() and attempts < n_attempts:
            self._logger.info('file lock found, waiting %.2f seconds %s', timeout, self.lockfile)
            time.sleep(timeout)
            attempts += 1

        # if the file still exists after 5 attempts, remove it as it's a job that went wrong
        if self.lockfile.exists():
            with open(self.lockfile, 'r') as fp:
                _contents = json.load(fp) if self.lockfile.stat().st_size else '<empty>'
                self._logger.debug('file lock contents: %s', _contents)
            if self.timeout_action == 'delete':
                self._logger.info('stale file lock found, deleting %s', self.lockfile)
                self.lockfile.unlink()
            else:
                raise TimeoutError(f'{self.lockfile} file lock timed out')

        # add in the lock file, add some metadata to ease debugging if one gets stuck
        with open(self.lockfile, 'w') as fp:
            json.dump(dict(datetime=datetime.utcnow().isoformat(), hostname=str(socket.gethostname)), fp)

    async def __aenter__(self):
        # if a lock file exists wait until timeout before removing
        try:
            await asyncio.wait_for(self._lock_check_async(), timeout=self.timeout)  # py3.11 use with asyncio.timeout
        except asyncio.TimeoutError as e:
            with open(self.lockfile, 'r') as fp:
                _contents = json.load(fp) if self.lockfile.stat().st_size else '<empty>'
                self._logger.debug('file lock contents: %s', _contents)
            if self.timeout_action == 'raise':
                raise e
            self._logger.info('stale file lock found, deleting %s', self.lockfile)
            self.lockfile.unlink()

        # add in the lock file, add some metadata to ease debugging if one gets stuck
        with open(self.lockfile, 'w') as fp:
            info = dict(datetime=datetime.utcnow().isoformat(), hostname=str(socket.gethostname), pid=os.getpid())
            json.dump(info, fp)

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.lockfile.unlink()

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self.lockfile.unlink()



import numpy as np
import unittest

from iblutil.spacer import Spacer


class TestSpacer(unittest.TestCase):

    def test_spacer(self):
        spacer = Spacer(dt_start=.02, dt_end=.4, n_pulses=8, tup=.05)
        np.testing.assert_equal(spacer.times.size, 15)
        sig = spacer.generate_template(fs=1000)
        ac = np.correlate(sig, sig, 'full') / np.sum(sig**2)
        # import matplotlib.pyplot as plt
        # plt.plot(ac)
        # plt.show()
        ac[sig.size - 100: sig.size + 100] = 0  # remove the main peak
        # the autocorrelation side lobes should be less than 30%
        assert np.max(ac) < .3

    def test_find_spacers(self):
        """Generates a fake signal with 2 spacers and finds them"""
        fs = 1000
        spacer = Spacer(dt_start=.02, dt_end=.4, n_pulses=8, tup=.05)
        start_times = [4.38, 96.58]
        template = spacer.generate_template(fs)
        signal = np.zeros(int(start_times[-1] * fs + template.size * 2))
        for start_time in start_times:
            signal[int(start_time * fs): int(start_time * fs) + template.size] = template
        spacer_times = spacer.find_spacers(signal, fs=fs)
        np.testing.assert_allclose(spacer_times, start_times)


class TestSpacersFromFronts(unittest.TestCase):
    """Tests for Spacer.find_spacers_from_fronts"""
    def setUp(self) -> None:
        self.spacer_1 = Spacer()
        self.spacer_2 = Spacer(n_pulses=12, tup=.1)

    def test_extract_spacers(self):
        """Test Spacer.find_spacers_from_fronts with usual spacer"""
        # Single spacer, nothing else
        Fs = 1000
        template = self.spacer_1.generate_template(fs=Fs)
        ind, val = self.to_fronts(template)
        t0 = 15.  # start at 15 seconds
        times = ind * 1 / Fs + t0
        fronts = {'times': times, 'polarities': val}
        t = self.spacer_1.find_spacers_from_fronts(fronts)
        self.assertEqual(1, len(t))
        np.testing.assert_allclose(t, t0, rtol=1e-3)

        # Add a second spacer with unrelated pulses in between
        signal_times, p = self._random_pulses(t0=t0 + 1 / Fs * template.size + np.random.rand())
        fronts['times'] = np.concatenate([times, signal_times, ind * 1 / Fs + t0 + signal_times[-1]])
        fronts['polarities'] = np.concatenate([val, p, val])

        # Expect 2 spacer times
        t = self.spacer_1.find_spacers_from_fronts(fronts)
        self.assertEqual(2, len(t))
        np.testing.assert_allclose(t, [t0, signal_times[-1] + t0], rtol=1e-3)

    def test_empty(self):
        """Test behaviour when a signal is passed in that doesn't contain any spacers."""
        # Make some random pulses
        times, polarities = self._random_pulses()
        fronts = {'times': times, 'polarities': polarities}
        self.assertEqual(0, self.spacer_1.find_spacers_from_fronts(fronts).size)

    def test_noise(self):
        """Test whether it can deal with noise in the spacer signal.

        NB: Currently any noise will result in no spacer detection.
        """
        Fs = 1000
        template = self.spacer_1.generate_template(fs=Fs)
        ind, val = self.to_fronts(template)
        t0 = 15.  # start at 15 seconds
        times = ind * 1 / Fs + t0

        # Add some noise to the signal
        noise_times, noise_p = self._random_pulses(t0, 4, pw=.005, t_max=times[-1] - t0)
        fronts = {
            'times': np.sort(np.concatenate([times, noise_times])),
            'polarities': np.concatenate([val, noise_p])
        }

        t = self.spacer_1.find_spacers_from_fronts(fronts)
        self.assertEqual(0, len(t))  # NB This could one day return 1
        np.testing.assert_allclose(t, t0, rtol=1e-3)

    def test_custom_spacer_extract(self):
        """Test is different spacer parameters"""
        Fs = 1200  # a non-default sampling rate
        template = self.spacer_2.generate_template(fs=Fs)
        ind, val = self.to_fronts(template)
        t0 = 15.  # start at 15 seconds
        fronts = {'times': ind * 1 / Fs + t0, 'polarities': val}
        t = self.spacer_2.find_spacers_from_fronts(fronts, fs=Fs)
        self.assertEqual(1, len(t))
        np.testing.assert_allclose(t, t0, rtol=1e-3)

    @staticmethod
    def _random_pulses(t0=0., n=50, pw=None, t_max=200):
        if not pw:
            times = np.sort(np.random.randint(0, t_max, n) + np.random.rand(n)) + t0
        else:
            times = np.random.randint(0, t_max, n) + np.random.rand(n) + t0
            times = np.sort(np.concatenate([times, times + pw]))
        polarities = np.ones_like(times)
        polarities[1::2] = -1.
        return times, polarities

    @staticmethod
    def to_fronts(signal):
        """
        Detects Rising and Falling edges of a voltage signal, returns indices and polarities.

        Parameters
        ----------
        signal : numpy.array
            Array on which to compute RMS

        Returns
        -------
        numpy.array
            Array of indices
        numpy.array
            Array of polarities {1, -1}

        """
        d = np.diff(signal)
        ind = np.array(np.where(np.abs(d) >= 1))
        sign = d[tuple(ind)]
        ind[-1] += 1
        return (ind[0], sign) if len(ind) == 1 else (ind, sign)



import unittest
from unittest import mock
import uuid
import tempfile
import os
from pathlib import Path
import json
import asyncio

import numpy as np

from iblutil.io.parquet import uuid2np, np2uuid, np2str, str2np
from iblutil.io import params
import iblutil.io.jsonable as jsonable
from iblutil.numerical import intersect2d, ismember2d, ismember


class TestParquet(unittest.TestCase):

    def test_uuids_conversions(self):
        str_uuid = 'a3df91c8-52a6-4afa-957b-3479a7d0897c'
        one_np_uuid = np.array([-411333541468446813, 8973933150224022421])
        two_np_uuid = np.tile(one_np_uuid, [2, 1])
        # array gives a list
        self.assertTrue(all(map(lambda x: x == str_uuid, np2str(two_np_uuid))))
        # single uuid gives a string
        self.assertTrue(np2str(one_np_uuid) == str_uuid)
        # list uuids with some None entries
        uuid_list = ['bc74f49f33ec0f7545ebc03f0490bdf6', 'c5779e6d02ae6d1d6772df40a1a94243',
                     None, '643371c81724378d34e04a60ef8769f4']
        assert np.all(str2np(uuid_list)[2, :] == 0)

    def test_uuids_intersections(self):
        ntotal = 500
        nsub = 17
        nadd = 3

        eids = uuid2np([uuid.uuid4() for _ in range(ntotal)])

        np.random.seed(42)
        isel = np.floor(np.argsort(np.random.random(nsub)) / nsub * ntotal).astype(np.int16)
        sids = np.r_[eids[isel, :], uuid2np([uuid.uuid4() for _ in range(nadd)])]
        np.random.shuffle(sids)

        # check the intersection
        v, i0, i1 = intersect2d(eids, sids)
        assert np.all(eids[i0, :] == sids[i1, :])
        assert np.all(np.sort(isel) == np.sort(i0))

        v_, i0_, i1_ = np.intersect1d(eids[:, 0], sids[:, 0], return_indices=True)
        assert np.setxor1d(v_, v[:, 0]).size == 0
        assert np.setxor1d(i0, i0_).size == 0
        assert np.setxor1d(i1, i1_).size == 0

        for a, b in zip(ismember2d(sids, eids), ismember(sids[:, 0], eids[:, 0])):
            assert np.all(a == b)

        # check conversion to numpy back and forth
        uuids = [uuid.uuid4() for _ in np.arange(4)]
        np_uuids = uuid2np(uuids)
        assert np2uuid(np_uuids) == uuids


class TestParams(unittest.TestCase):

    @mock.patch('sys.platform', 'linux')
    def test_set_hidden(self):
        with tempfile.TemporaryDirectory() as td:
            file = Path(td).joinpath('file')
            file.touch()
            hidden_file = params.set_hidden(file, True)
            self.assertFalse(file.exists())
            self.assertTrue(hidden_file.exists())
            self.assertEqual(hidden_file.name, '.file')

            params.set_hidden(hidden_file, False)
            self.assertFalse(hidden_file.exists())
            self.assertTrue(file.exists())


class TestFileLock(unittest.IsolatedAsyncioTestCase):
    tmp = None

    @classmethod
    def setUpClass(cls):
        tmp = tempfile.TemporaryDirectory()
        cls.tmp = Path(tmp.name)
        cls.addClassCleanup(tmp.cleanup)

    def setUp(self):
        self.file = self.tmp / 'foo.bar'
        self.addCleanup(self.file.unlink, missing_ok=True)
        self.lock_file = self.file.with_suffix('.lock')
        self.addCleanup(self.lock_file.unlink, missing_ok=True)

    @mock.patch('iblutil.io.params.time.sleep')
    def test_file_lock_sync(self, sleep_mock):
        """Test synchronous FileLock context manager."""
        # Check input validation
        self.assertRaises(ValueError, params.FileLock, self.file, timeout_action='foo')

        # Check behaviour when lock file doesn't exist (i.e. no other process writing to file)
        assert not self.lock_file.exists()
        with params.FileLock(self.file, timeout_action='raise'):
            self.assertTrue(self.lock_file.exists(), 'Failed to create lock file')
        self.assertFalse(self.lock_file.exists(), 'Failed to remove lock file upon exit of context manager')
        sleep_mock.assert_not_called()  # no file present so no need to sleep

        # Check behaviour when lock file present and not removed by other process
        self.lock_file.touch()
        assert self.lock_file.exists()
        lock = params.FileLock(self.file, timeout_action='raise')
        with self.assertLogs('iblutil.io.params', 10) as lg:
            self.assertRaises(TimeoutError, lock.__enter__)
        msg = next((x.getMessage() for x in lg.records if x.levelno == 10), None)
        self.assertEqual('file lock contents: <empty>', msg)
        # should try 5 attempts by default; default total timeout is 10 seconds so should sleep 5x for 2 seconds each
        expected_attempts = 5
        sleep_mock.assert_called_with(2)
        self.assertEqual(expected_attempts, sleep_mock.call_count)
        self.assertEqual(expected_attempts, len([x for x in lg.records if x.levelno == 20]))
        msg = next(x.getMessage() for x in lg.records if x.levelno == 20)
        self.assertRegex(msg, 'file lock found, waiting 2.00 seconds')

        # Check delete timeout action
        assert self.lock_file.exists()
        with self.assertLogs('iblutil.io.params', 10) as lg, \
                params.FileLock(self.file, timeout_action='delete'):
            # Should have replaced empty lock file with timestamped one
            self.assertTrue(self.lock_file.exists())
            with open(self.lock_file, 'r') as fp:
                lock_info = json.load(fp)
            self.assertCountEqual(('datetime', 'hostname'), lock_info)
        self.assertFalse(self.lock_file.exists(), 'Failed to remove lock file upon exit of context manager')
        self.assertRegex(lg.records[-1].getMessage(), 'stale file lock found, deleting')

    async def _mock(self, obj):
        """
        Add side effect to mock object that awaits a future.

        This is required because async lambdas are not supported.

        Parameters
        ----------
        obj : unittest.mock.AsyncMock
            An asynchronous mock object to install side effect for.

        Returns
        -------
        asyncio.Future
            A future awaited by input mock object.
        """
        fut = asyncio.get_event_loop().create_future()
        self.addCleanup(fut.cancel)

        async def wait(_):
            return await fut

        obj.side_effect = wait
        return fut

    @mock.patch('iblutil.io.params.asyncio.sleep')
    async def test_file_lock_async(self, sleep_mock):
        """Test asynchronous FileLock context manager."""
        # Check behaviour when lock file doesn't exist (i.e. no other process writing to file)
        assert not self.lock_file.exists()
        async with params.FileLock(self.file, timeout_action='raise'):
            self.assertTrue(self.lock_file.exists(), 'Failed to create lock file')
        self.assertFalse(self.lock_file.exists(), 'Failed to remove lock file upon exit of context manager')
        sleep_mock.assert_not_called()  # no file present so no need to sleep

        # Check behaviour when lock file present and not removed by other process
        self.lock_file.touch()
        assert self.lock_file.exists()
        lock = params.FileLock(self.file, timeout=1e-3, timeout_action='raise')
        # The loop that checks the lock file is too fast when async.sleep is mocked so adding a side
        # effect that awaits a future that's never set allows the timeout code to execute.
        await self._mock(sleep_mock)

        with self.assertLogs('iblutil.io.params', 10) as lg, self.assertRaises(asyncio.TimeoutError):
            await lock.__aenter__()
            # fut = asyncio.get_running_loop().create_future()
            # with mock.patch.object(lock, '_lock_check_async', return_value) as m:

            # async with params.FileLock(self.file, timeout=1e-3, timeout_action='raise') as lock:
            #     ...
        sleep_mock.assert_awaited_with(lock._async_poll_freq)
        msg = next((x.getMessage() for x in lg.records if x.levelno == 10), None)
        self.assertEqual('file lock contents: <empty>', msg)

        # Check remove timeout action
        assert self.lock_file.exists()
        await self._mock(sleep_mock)
        with self.assertLogs('iblutil.io.params', 10) as lg:
            async with params.FileLock(self.file, timeout=1e-5, timeout_action='delete'):
                # Should have replaced empty lock file with timestamped one
                self.assertTrue(self.lock_file.exists())
                with open(self.lock_file, 'r') as fp:
                    lock_info = json.load(fp)
            self.assertCountEqual(('datetime', 'hostname', 'pid'), lock_info)
        self.assertFalse(self.lock_file.exists(), 'Failed to remove lock file upon exit of context manager')


class TestsJsonable(unittest.TestCase):
    def setUp(self) -> None:
        self.tfile = tempfile.NamedTemporaryFile(delete=False)

    def testReadWrite(self):
        data = [{'a': 'thisisa', 'b': 1, 'c': [1, 2, 3]},
                {'a': 'thisisb', 'b': 2, 'c': [2, 3, 4]}]
        jsonable.write(self.tfile.name, data)
        data2 = jsonable.read(self.tfile.name)
        self.assertEqual(data, data2)
        jsonable.append(self.tfile.name, data)
        data3 = jsonable.read(self.tfile.name)
        self.assertEqual(data + data, data3)

    def tearDown(self) -> None:
        self.tfile.close()
        os.unlink(self.tfile.name)


if __name__ == '__main__':
    unittest.main(exit=False, verbosity=2)



import unittest
import uuid

import iblutil.numerical as num
import numpy as np
import pandas as pd


class HashUUIDs(unittest.TestCase):

    def test_hash_uuids(self):
        the_hash = 'fc898a58172960bcc2d49577e17d34e0e59ded74567ee236533c0757f016262f'
        uuids = [
            'c5cabbda-1f74-4168-ad5e-7c5a2f533d9f',
            '5b5176f2-228f-46b5-85c8-205cf9d90a53',
            '19930b44-2f8a-44e1-a8ce-46184f7334cb'
        ]
        self.assertEqual(the_hash, num.hash_uuids(uuids))
        self.assertEqual(the_hash, num.hash_uuids(np.array(uuids)))
        self.assertEqual(the_hash, num.hash_uuids(pd.Series(uuids)))
        self.assertEqual(the_hash, num.hash_uuids([uuid.UUID(hex=uid) for uid in uuids]))


class TestRcoeff(unittest.TestCase):
    def test_rcoeff(self):
        x = np.random.rand(2, 1000)
        y = x[0, :]
        r = np.corrcoef(x[1, :], y)
        assert num.rcoeff(x[0, :], y) == 1
        assert np.isclose(num.rcoeff(x[1, :], y), r[1, 0])
        assert np.all(np.isclose(num.rcoeff(x, y), r[0, :]))
        assert np.all(np.isclose(num.rcoeff(y, x), r[0, :]))


class TestBetweeenSorted(unittest.TestCase):
    def test_between_sorted_single_time(self):
        # test single time, falling right on edges
        t = np.arange(100)
        bounds = [10, 25]
        ind = num.between_sorted(t, bounds)
        assert np.all(
            t[ind] == np.arange(int(np.ceil(bounds[0])), int(np.floor(bounds[1] + 1)))
        )
        # test single time in between edges
        bounds = [10.4, 25.2]
        ind = num.between_sorted(t, bounds)
        assert np.all(t[ind] == np.arange(np.ceil(bounds[0]), np.floor(bounds[1]) + 1))
        # out of bounds
        ind = num.between_sorted(t, [-5, 800])
        assert np.all(ind)

    def test_between_sorted_multiple_times(self):
        t = np.arange(100)
        # non overlapping ranges
        bounds = np.array([[10.4, 25.2], [67.2, 86.4]])
        ind_ = np.logical_or(
            num.between_sorted(t, bounds[0]), num.between_sorted(t, bounds[1])
        )
        ind = num.between_sorted(t, bounds)
        assert np.all(ind == ind_)
        # overlapping ranges
        bounds = np.array([[10.4, 83.2], [67.2, 86.4]])
        ind_ = np.logical_or(
            num.between_sorted(t, bounds[0]), num.between_sorted(t, bounds[1])
        )
        ind = num.between_sorted(t, bounds)
        assert np.all(ind == ind_)
        # one range contains the other
        bounds = np.array([[10.4, 83.2], [34, 78]])
        ind_ = np.logical_or(
            num.between_sorted(t, bounds[0]), num.between_sorted(t, bounds[1])
        )
        ind = num.between_sorted(t, bounds)
        assert np.all(ind == ind_)
        # case when one range starts exactly where another stops
        bounds = np.array([[10.4, 83.2], [83.2, 84]])
        ind = num.between_sorted(t, bounds)
        ind_ = np.logical_or(
            num.between_sorted(t, bounds[0]), num.between_sorted(t, bounds[1])
        )
        assert np.all(ind == ind_)

    def test_between_sorted_out_of_range(self):
        # np searchsorted was returning out of range index when the start time
        # was greater than the max or the end time lower than the min
        bounds = np.array([[-10.4, -3], [10.4, 20], [34, 78]])
        array = np.arange(30)
        assert np.sum(num.between_sorted(array, bounds)) == 10


class TestIsmember(unittest.TestCase):
    def test_ismember2d(self):
        b = np.reshape([0, 0, 0, 1, 1, 1], [3, 2])
        locb = np.array([0, 1, 0, 2, 2, 1])
        lia = np.array([True, True, True, True, True, True, False, False])
        a = np.r_[b[locb, :], np.array([[2, 1], [1, 2]])]
        lia_, locb_ = num.ismember2d(a, b)
        assert np.all(lia == lia_) & np.all(locb == locb_)

    def test_ismember2d_uuids(self):
        nb = 20
        na = 500
        np.random.seed(42)
        a = np.random.randint(0, nb + 3, na)
        b = np.arange(nb)
        lia, locb = num.ismember(a, b)
        bb = np.random.randint(
            low=np.iinfo(np.int64).min,
            high=np.iinfo(np.int64).max,
            size=(nb, 2),
            dtype=np.int64,
        )
        aa = np.zeros((na, 2), dtype=np.int64)
        aa[lia, :] = bb[locb, :]
        lia_, locb_ = num.ismember2d(aa, bb)
        assert np.all(lia == lia_) & np.all(locb == locb_)
        bb[:, 0] = 0
        aa[:, 0] = 0
        # if the first column is equal, the distinction is to be made on the second\
        assert np.unique(bb[:, 1]).size == nb
        lia_, locb_ = num.ismember2d(aa, bb)
        assert np.all(lia == lia_) & np.all(locb == locb_)

    def test_ismember(self):
        def _check_ismember(a, b, lia_, locb_):
            lia, locb = num.ismember(a, b)
            self.assertTrue(np.all(a[lia] == b[locb]))
            self.assertTrue(np.all(lia_ == lia))
            self.assertTrue(np.all(locb_ == locb))

        b = np.array([0, 1, 3, 4, 4])
        a = np.array([1, 4, 5, 4])
        lia_ = np.array([True, True, False, True])
        locb_ = np.array([1, 3, 3])
        _check_ismember(a, b, lia_, locb_)

        b = np.array([0, 4, 3, 1, 4])
        a = np.array([1, 4, 5, 4])
        lia_ = np.array([True, True, False, True])
        locb_ = np.array([3, 1, 1])
        _check_ismember(a, b, lia_, locb_)

        b = np.array([0, 1, 3, 4])
        a = np.array([1, 4, 5])
        lia_ = np.array([True, True, False])
        locb_ = np.array([1, 3])
        _check_ismember(a, b, lia_, locb_)


class TestWithinRanges(unittest.TestCase):
    def test_within_ranges(self):
        verifiable = num.within_ranges(np.arange(11), [(1, 2), (5, 8)])
        expected = np.array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0], dtype=int)
        np.testing.assert_array_equal(verifiable, expected)

        # Matrix mode
        ranges = np.array([[1, 2], [5, 8]])
        verifiable = num.within_ranges(
            np.arange(10) + 1, ranges, labels=np.array([0, 1]), mode="matrix"
        )
        expected = np.array(
            [[1, 1, 0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 1, 1, 1, 0, 0]], dtype=int
        )
        np.testing.assert_array_equal(verifiable, expected)

        # Test overlap
        verifiable = num.within_ranges(
            np.arange(11), [(1, 2), (5, 8), (4, 6)], labels=[0, 1, 1], mode="matrix"
        )
        expected = np.array(
            [[0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 2, 2, 1, 1, 0, 0]],
            dtype=int,
        )
        np.testing.assert_array_equal(verifiable, expected)

        # Vector mode
        verifiable = num.within_ranges(
            np.arange(10) + 1, ranges, np.array([3, 1]), mode="vector"
        )
        expected = np.array([3, 3, 0, 0, 1, 1, 1, 1, 0, 0], dtype=int)
        np.testing.assert_array_equal(verifiable, expected)

        # Boolean
        verifiable = num.within_ranges(
            np.arange(11), [(1, 2), (5, 8), (4, 6)], dtype=bool
        )
        expected = np.array(
            [False, True, True, False, True, True, True, True, True, False, False]
        )
        np.testing.assert_array_equal(verifiable, expected)

        # Edge cases
        verifiable = num.within_ranges(np.arange(11), [])
        expected = np.zeros(11, dtype=int)
        np.testing.assert_array_equal(verifiable, expected)

        with self.assertRaises(ValueError):
            num.within_ranges(np.arange(11), [(1, 2)], mode="array")


class TestBincount2D(unittest.TestCase):
    def test_bincount_2d(self):
        # first test simple with indices
        x = np.array([0, 1, 1, 2, 2, 3, 3, 3])
        y = np.array([3, 2, 2, 1, 1, 0, 0, 0])
        r, xscale, yscale = num.bincount2D(x, y, xbin=1, ybin=1)
        r_ = np.zeros_like(r)
        # sometimes life would have been simpler in c:
        for ix, iy in zip(x, y):
            r_[iy, ix] += 1
        self.assertTrue(np.all(np.equal(r_, r)))
        # test with negative values
        y = np.array([3, 2, 2, 1, 1, 0, 0, 0]) - 5
        r, xscale, yscale = num.bincount2D(x, y, xbin=1, ybin=1)
        self.assertTrue(np.all(np.equal(r_, r)))
        # test unequal bins
        r, xscale, yscale = num.bincount2D(x / 2, y / 2, xbin=1, ybin=2)
        r_ = np.zeros_like(r)
        for ix, iy in zip(np.floor(x / 2), np.floor((y / 2 + 2.5) / 2)):
            r_[int(iy), int(ix)] += 1
        self.assertTrue(np.all(r_ == r))
        # test with weights
        w = np.ones_like(x) * 2
        r, xscale, yscale = num.bincount2D(x / 2, y / 2, xbin=1, ybin=2, weights=w)
        self.assertTrue(np.all(r_ * 2 == r))
        # test aggregation instead of binning
        x = np.array([0, 1, 1, 2, 2, 4, 4, 4])
        y = np.array([4, 2, 2, 1, 1, 0, 0, 0])
        r, xscale, yscale = num.bincount2D(x, y)
        self.assertTrue(
            np.all(xscale == yscale) and np.all(xscale == np.array([0, 1, 2, 4]))
        )
        # test aggregation on a fixed scale
        r, xscale, yscale = num.bincount2D(
            x + 10, y + 10, xbin=np.arange(5) + 10, ybin=np.arange(3) + 10
        )
        self.assertTrue(np.all(xscale == np.arange(5) + 10))
        self.assertTrue(np.all(yscale == np.arange(3) + 10))
        self.assertTrue(np.all(r.shape == (3, 5)))



import unittest
import types
from pathlib import Path
import tempfile
import logging
from unittest.mock import patch

import numpy as np

from iblutil import util


class TestBunch(unittest.TestCase):

    def test_copy(self):
        """Test Bunch.copy method."""
        # Expect shallow copy by default
        bunch_one = util.Bunch({'test': np.arange(5)})
        bunch_two = bunch_one.copy()
        self.assertIsNot(bunch_one, bunch_two)
        bunch_two['test'][0] = 5
        self.assertEqual(5, bunch_one['test'][0])
        # Expect deep copy
        bunch_one = util.Bunch({'test': np.arange(5)})
        bunch_two = bunch_one.copy(deep=True)
        self.assertIsNot(bunch_one, bunch_two)
        bunch_two['test'][0] = 5
        self.assertNotEqual(5, bunch_one['test'][0])

    def test_sync(self):
        """
        This test is just to document current use in libraries in case of refactoring
        """
        sd = util.Bunch({'label': 'toto', 'ap': None, 'lf': 8})
        self.assertTrue(sd['label'] is sd.label)
        self.assertTrue(sd['ap'] is sd.ap)
        self.assertTrue(sd['lf'] is sd.lf)

    def test_bunch_io(self):
        a = np.random.rand(50, 1)
        b = np.random.rand(50, 1)
        abunch = util.Bunch({'a': a, 'b': b})

        with tempfile.TemporaryDirectory() as td:
            npz_file = Path(td).joinpath('test_bunch.npz')
            abunch.save(npz_file)
            another_bunch = util.Bunch.load(npz_file)
            [self.assertTrue(np.all(abunch[k]) == np.all(another_bunch[k])) for k in abunch]
            npz_filec = Path(td).joinpath('test_bunch_comp.npz')
            abunch.save(npz_filec, compress=True)
            another_bunch = util.Bunch.load(npz_filec)
            [self.assertTrue(np.all(abunch[k]) == np.all(another_bunch[k])) for k in abunch]
            with self.assertRaises(FileNotFoundError):
                util.Bunch.load(Path(td) / 'fake.npz')


class TestFlatten(unittest.TestCase):

    def test_flatten(self):
        x = (1, 2, 3, [1, 2], 'string', 0.1, {1: None}, [[1, 2, 3], {1: 1}, 1])
        self.assertEqual(util._iflatten(x), util.flatten(x))
        self.assertEqual(util.flatten(x)[:5], [1, 2, 3, 1, 2])
        self.assertEqual(list(util._gflatten(x)), list(util.flatten(x, generator=True)))
        self.assertIsInstance(util.flatten(x, generator=True), types.GeneratorType)


class TestRangeStr(unittest.TestCase):

    def test_range_str(self):
        x = [1, 2, 3, 4, 5, 6, 7, 8, 12, 17]
        self.assertEqual(util.range_str(x), '1-8, 12 & 17')

        x = [0, 6, 7, 10, 11, 12, 30, 30]
        self.assertEqual(util.range_str(x), '0, 6-7, 10-12 & 30')

        self.assertEqual(util.range_str([]), '')


class TestLogger(unittest.TestCase):
    log_name = '_foobar'

    def test_no_duplicates(self):
        log = util.setup_logger('gnagna')
        self.assertEqual(1, len(log.handlers))
        log = util.setup_logger('gnagna')
        self.assertEqual(1, len(log.handlers))

    def test_file_handler_setup(self):
        # NB: this doesn't work with a context manager, the handlers get all confused
        # with the fake file object
        with tempfile.TemporaryDirectory() as tn:
            file_log = Path(tn).joinpath('log.txt')
            log = util.setup_logger('tutu', file=file_log, no_color=True, level=20)
            log.info('toto')
            # the purpose of the test is to test that the logger/handler has not been
            # duplicated so after 2 calls we expect 2 lines
            log = util.setup_logger('tutu', file=file_log, level=20)
            log.info('tata')
            while True:
                handlers = log.handlers
                if len(handlers) == 0:
                    break
                handlers[0].close()
                log.removeHandler(handlers[0])
            with open(file_log) as fp:
                lines = fp.readlines()
            self.assertEqual(3, len(lines))

    def test_file_handler_stand_alone(self):
        """Test for ibllib.misc.log_to_file"""
        log_path = Path.home().joinpath('.ibl_logs', self.log_name)
        log_path.unlink(missing_ok=True)
        test_log = util.log_to_file(filename=self.log_name, log=self.log_name)
        test_log.level = 20
        test_log.info('foobar')

        # Should have created a log file and written to it
        self.assertTrue(log_path.exists())
        with open(log_path, 'r') as f:
            logged = f.read()
        self.assertIn('foobar', logged)

    def test_file_handler_log_input(self):
        """Test ibllib.misc.log_to_file accepts log object as input"""
        # Should word with log as input
        test_log = logging.getLogger(self.log_name)
        test_log.level = 20
        util.log_to_file(test_log).info('hello world!')
        log_path = Path.home().joinpath('.ibl_logs', self.log_name)
        self.assertTrue(log_path.exists())
        with open(log_path, 'r') as f:
            logged = f.read()
        self.assertIn('hello world', logged)

    def tearDown(self) -> None:
        # Before we can delete the test log file we must close the file handler
        test_log = logging.getLogger(self.log_name)
        for handler in test_log.handlers:
            handler.close()
            test_log.removeHandler(handler)
        Path.home().joinpath('.ibl_logs', self.log_name).unlink(missing_ok=True)


class TestRrmdir(unittest.TestCase):

    def test_rrmdir(self):
        with self.assertRaises(FileNotFoundError):
            util.rrmdir(Path('/fantasy/path'))
        with tempfile.TemporaryDirectory() as tempdir:
            # nested folders
            folder_level_3 = Path(tempdir).joinpath('folder_level_3')
            folder_level_2 = folder_level_3.joinpath('folder_level_2')
            folder_level_1 = folder_level_2.joinpath('folder_level_1')
            folder_level_0 = folder_level_1.joinpath('folder_level_0')

            # default level = 0, folder contains file - nothing should happen
            folder_level_0.mkdir(parents=True)
            (file := folder_level_0.joinpath('file')).touch()
            self.assertEqual([], util.rrmdir(folder_level_0))
            self.assertTrue(file.exists())

            # default level = 0, folder and all parents are empty
            file.unlink()
            self.assertEqual([folder_level_0], util.rrmdir(folder_level_0))
            self.assertFalse(folder_level_0.exists())
            self.assertTrue(folder_level_1.exists())

            # remove empty folders to level 2
            folder_level_0.mkdir(parents=True)
            removed = util.rrmdir(folder_level_0, levels=2)
            self.assertEqual([folder_level_0, folder_level_1, folder_level_2], removed)
            self.assertFalse(folder_level_2.exists())
            self.assertTrue(folder_level_3.exists())

            # remove empty folders to level 3, with a file in level 2
            folder_level_0.mkdir(parents=True)
            (file := folder_level_2.joinpath('file')).touch()
            removed = util.rrmdir(folder_level_0, levels=3)
            self.assertEqual(removed, [folder_level_0, folder_level_1])
            self.assertFalse(folder_level_1.exists())
            self.assertTrue(file.exists())


class TestDirSize(unittest.TestCase):

    def test_dir_size(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            dir1 = Path(temp_dir)
            dir2 = Path(dir1).joinpath('sub_dir')
            dir2.mkdir()
            file1 = dir1.joinpath('file1')
            file2 = dir2.joinpath('file2')
            file3 = dir2.joinpath('file3')
            with open(file1, 'w') as f1, open(file2, 'w') as f2, open(file3, 'w') as f3:
                f1.write('Old pond')
                f2.write('A frog jumps in')
                f3.write('The sound of water')
            symlink = dir2.joinpath('symlink_file')
            symlink.symlink_to(file1)
            expected = file1.stat().st_size + file2.stat().st_size + file3.stat().st_size
            self.assertEqual(util.dir_size(str(dir1)), expected)
            self.assertEqual(util.dir_size(dir1), expected)
            self.assertEqual(util.dir_size(dir1, True), expected + file1.stat().st_size)


class TestGetMac(unittest.TestCase):

    def test_get_mac(self):
        with patch('iblutil.util.uuid.getnode', return_value=205452675710958):
            self.assertEqual(util.get_mac(), 'BA-DB-AD-C0-FF-EE')


class TestEnsureList(unittest.TestCase):
    """Test ensure_list function."""

    def test_ensure_list(self):
        """Test ensure_list function."""
        x = [1, 2, 3]
        self.assertIs(x, util.ensure_list(x))
        x = tuple(x)
        self.assertIs(x, util.ensure_list(x))
        # Shouldn't iterate over strings or dicts
        x = '123'
        self.assertEqual([x], util.ensure_list(x))
        x = {'1': 1, '2': 2, '3': 3}
        self.assertEqual([x], util.ensure_list(x))
        # Check exclude_type param behaviour
        x = np.array([1, 2, 3])
        self.assertIs(x, util.ensure_list(x))
        self.assertEqual([x], util.ensure_list(x, exclude_type=(np.ndarray)))


if __name__ == '__main__':
    unittest.main(exit=False)



import sys
import asyncio
import logging
import unittest
from unittest import mock
import ipaddress
import socket
from packaging.version import Version
from datetime import date

from iblutil.io.net import base, app

ver = (getattr(sys.version_info, v) for v in ('major', 'minor', 'micro'))
ver = Version('.'.join(map(str, ver)))


class TestBase(unittest.IsolatedAsyncioTestCase):
    """Test for base network utils.

    NB: This requires internet access.
    """
    def test_parse_uri(self):
        """Tests for parse_uri, validate_ip and hostname2ip"""
        expected = 'udp://192.168.0.1:9999'
        uri = base.validate_uri(expected)
        self.assertEqual(expected, uri)
        self.assertEqual(expected, base.validate_uri(uri[6:]))
        self.assertEqual(expected.replace('udp', 'ws'), base.validate_uri(uri[6:], default_proc='ws'))
        self.assertEqual(expected, base.validate_uri(uri[:-5], default_port=9999))
        uri = base.validate_uri(ipaddress.ip_address('192.168.0.1'), default_port=9999)
        self.assertEqual(expected, uri)
        self.assertEqual('udp://foobar:11001', base.validate_uri('foobar', resolve_host=False))
        # Check IP resolved
        uri = base.validate_uri('http://google.com:80', resolve_host=True)
        expected = (ipaddress.IPv4Address, ipaddress.IPv6Address)
        self.assertIsInstance(ipaddress.ip_address(uri[7:-3]), expected)
        # Check validations
        validations = {'ip': '256.168.0.0000', 'hostname': 'foo@bar$', 'port': 'foobar:00'}
        for subtest, to_validate in validations.items():
            with self.subTest(**{subtest: to_validate}):
                with self.assertRaises(ValueError):
                    base.validate_uri(to_validate, resolve_host=False)
        with self.assertRaises(ValueError):
            base.validate_uri(' ', resolve_host=True)
        with self.assertRaises(TypeError):
            base.validate_uri(b'localhost')

    def test_external_ip(self):
        """Test for external_ip"""
        self.assertFalse(ipaddress.ip_address(base.external_ip()).is_private)

    def test_ExpMessage(self):
        """Test for ExpMessage.validate method."""
        # Check identity
        msg = base.ExpMessage.validate(base.ExpMessage.EXPINFO)
        self.assertIs(msg, base.ExpMessage.EXPINFO)

        # Check integer input
        msg = base.ExpMessage.validate(int(base.ExpMessage.EXPCLEANUP))
        self.assertIs(msg, base.ExpMessage.EXPCLEANUP)

        # Check string input
        msg = base.ExpMessage.validate(' expstatus')
        self.assertIs(msg, base.ExpMessage.EXPSTATUS)

        # Check errors
        with self.assertRaises(TypeError):
            base.ExpMessage.validate(b'EXPSTART')
        with self.assertRaises(ValueError):
            base.ExpMessage.validate('EXPSTOP')

        # Test allow_bitwise kwarg
        event = base.ExpMessage.any()
        self.assertIs(base.ExpMessage.validate(event), event)
        with self.assertRaises(ValueError):
            base.ExpMessage.validate(event, allow_bitwise=False)

    def test_encode(self):
        """Tests for iblutil.io.net.base.Communicator.encode"""
        message = [None, 21, 'message']
        encoded = base.Communicator.encode(message)
        self.assertEqual(encoded, b'[null, 21, "message"]')
        self.assertEqual(base.Communicator.encode(encoded), b'[null, 21, "message"]')

    def test_decode(self):
        """Tests for iblutil.io.net.base.Communicator.decode"""
        data = b'[null, 21, "message"]'
        decoded = base.Communicator.decode(data)
        self.assertEqual(decoded, [None, 21, 'message'])
        with self.assertWarns(Warning):
            decoded = base.Communicator.decode(data + b'"')
            self.assertEqual(decoded, '[null, 21, "message"]"')

    async def test_is_success(self):
        """Tests for iblutil.io.net.base.is_success function."""
        # Expect True only when set_result has been called.
        fut = asyncio.get_event_loop().create_future()
        self.assertFalse(base.is_success(fut))
        fut.set_result(None)
        self.assertTrue(base.is_success(fut))
        fut = asyncio.get_event_loop().create_future()
        fut.cancel()
        self.assertFalse(base.is_success(fut))
        fut = asyncio.get_event_loop().create_future()
        fut.set_exception(RuntimeError)
        self.assertFalse(base.is_success(fut))


@unittest.skipIf(ver < Version('3.9'), 'only version 3.9 or later supported')
class TestUDP(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.server = await app.EchoProtocol.server('localhost', name='server')
        self.client = await app.EchoProtocol.client('localhost', name='client')

    async def test_start(self):
        """Tests confirmed send via start command."""
        # Check socket type
        self.assertIs(self.server._socket.type, app.socket.SOCK_DGRAM)
        self.assertIs(self.client._socket.type, app.socket.SOCK_DGRAM)

        spy = mock.MagicMock()
        self.server.assign_callback('expstart', spy)
        exp_ref = '2022-01-01_1_subject'
        with self.assertLogs(self.server.logger, logging.INFO) as log:
            await self.client.start(exp_ref)
            self.assertIn(f'Received \'[{base.ExpMessage.EXPSTART}, "{exp_ref}", null]', log.records[-1].message)
        spy.assert_called_with([exp_ref, None], (self.client._socket.getsockname()))

    async def test_callback_error(self):
        """Tests behaviour when callback raises exception."""
        callback = mock.MagicMock()
        callback.side_effect = ValueError('Callback failed')

        self.server.assign_callback('expinit', callback)
        task = asyncio.create_task(self.client.on_event(0))
        with self.assertLogs(self.server.logger, logging.ERROR) as log:
            base.Communicator._receive(self.server, b'[1, null]', self.client._socket.getsockname())
            self.assertEqual(1, len(log.records))
            self.assertIn('Callback failed', log.records[-1].message)

        # Check error propagated back to client
        with self.assertLogs(self.client.logger, logging.ERROR) as log:
            (err, evt), _ = await task
            self.assertEqual(1, len(log.records))
            self.assertIn('Callback failed', log.records[-1].message)
        self.assertEqual(base.ExpMessage.EXPINIT.value, evt)
        self.assertIn('Callback failed', err)

        # Check behaviour when future already done
        fut = asyncio.get_running_loop().create_future()
        self.server.assign_callback('EXPSTART', fut)
        fut.set_result(True)
        task = asyncio.create_task(self.client.on_event(2))
        with self.assertLogs(self.server.logger, logging.WARNING) as log:
            base.Communicator._receive(self.server, b'[2, null]', self.client._socket.getsockname())
            self.assertEqual(1, len(log.records))
            self.assertRegex(log.records[-1].getMessage(), 'Future .+ already resolved')

    async def test_on_event(self):
        """Test on_event method as well as init, stop, etc."""
        # INIT
        task = asyncio.create_task(self.server.on_event('expinit'))
        await self.client.init(42)
        actual, _ = await task
        self.assertEqual([42], actual)

        # CLEANUP
        task = asyncio.create_task(self.server.on_event(base.ExpMessage.EXPCLEANUP))
        await self.client.cleanup(8)
        actual, _ = await task
        self.assertEqual([8], actual)

        # STOP
        task = asyncio.create_task(self.server.on_event('EXPEND'))
        await self.client.stop('foo')
        actual, _ = await task
        self.assertEqual(['foo'], actual)

        # INTERRUPT
        task = asyncio.create_task(self.server.on_event('expinterrupt'))
        await self.client.stop('foo', immediately=True)
        actual, _ = await task
        self.assertEqual(['foo'], actual)

        # START
        task = asyncio.create_task(self.server.on_event('expstart'))
        await self.client.start('2020-01-01_1_baz', {'foo': 'bar'})
        actual, _ = await task
        self.assertEqual(['2020-01-01_1_baz', {'foo': 'bar'}], actual)
        task = asyncio.create_task(self.server.on_event('expstart'))
        ref = {'subject': 'baz', 'date': date(2020, 1, 1), 'sequence': 1}
        await self.client.start(ref, {'foo': 'bar'})
        actual, _ = await task
        self.assertEqual(['2020-01-01_1_baz', {'foo': 'bar'}], actual)

        # STATUS
        task = asyncio.create_task(self.server.on_event('EXPSTATUS'))
        await self.client.status(base.ExpStatus.STOPPED)
        actual, _ = await task
        self.assertEqual([base.ExpStatus.STOPPED.value], actual)
        task = asyncio.create_task(self.server.on_event('EXPSTATUS'))
        await self.client.status('CONNECTED')
        actual, _ = await task
        self.assertEqual([base.ExpStatus.CONNECTED.value], actual)

        # INFO
        task = asyncio.create_task(self.server.on_event('expinfo'))
        await self.client.info(base.ExpStatus.RUNNING, {'foo': 'bar'})
        actual, _ = await task
        self.assertEqual([base.ExpStatus.RUNNING.value, {'foo': 'bar'}], actual)

    async def test_alyx(self):
        """Test iblutil.io.net.app.EchoProtocol.alyx method."""
        # Mock an AlyxClient instance that is logged in
        alyx = mock.MagicMock()
        alyx.is_logged_in = True
        alyx.base_url = 'https://alyx.website.net'
        alyx.user = 'foo'
        alyx._token = {'token': '4157aa522b855239cd05f4d23d40563aa0518359'}

        # When Alyx instance is logged in, expect the token to be broadcast
        task = asyncio.create_task(self.server.on_event('ALYX'))
        self.assertIsNone(await self.client.alyx(alyx), 'unexpected argument returned')
        actual, _ = await task  # wait for server process request
        expected = ['https://alyx.website.net', {'foo': {'token': '4157aa522b855239cd05f4d23d40563aa0518359'}}]
        self.assertEqual(expected, actual, 'client failed to send alyx token to server')

        # Client should request and return a token when Alyx instance is not logged in
        async def _req_callback():
            data, addr = await self.server.on_event('ALYX')
            self.assertEqual(['https://alyx.website.net', {}], data)
            await self.server.alyx(alyx, addr)

        # Mock an AlyxClient instance that is not logged in
        alyx_logged_out = mock.MagicMock()
        alyx_logged_out.is_logged_in = False
        alyx_logged_out.base_url = 'https://alyx.website.net'
        alyx_logged_out.user = alyx_logged_out._token = None

        task = asyncio.create_task(_req_callback())
        token = await self.client.alyx(alyx_logged_out)
        await task  # wait for server process request
        self.assertEqual(expected, token, 'failed to return requested token from server')

        # Try the same thing but with None instead of an Alyx instance
        async def _req_callback():
            data, addr = await self.server.on_event('ALYX')
            self.assertEqual([None, {}], data)
            await self.server.alyx(alyx, addr)

        task = asyncio.create_task(_req_callback())
        token = await self.client.alyx(None)
        await task  # wait for server process request
        self.assertEqual(expected, token)

    async def test_confirmed_send_validation(self):
        """Basic tests for iblutil.io.net.app.EchoProtocol.confirmed_send exception handling."""
        # Expect to raise in server role when no address provided
        with self.assertRaises(TypeError):
            await self.server.confirmed_send(None)
        # Expect to raise in client role when provided address does not match server URI
        with self.assertRaises(ValueError):
            await self.client.confirmed_send(None, addr=('localhost', self.client.port + 100))
        # Expect to raise when echo timeout is 0
        try:
            self.client.default_echo_timeout = 0
            with self.assertRaises(ValueError):
                await self.client.confirmed_send(None)
        finally:
            self.client.default_echo_timeout = app.EchoProtocol.default_echo_timeout
        # Expect timeout arg to override default echo timeout, expect error raised on timeout
        assert self.client.is_connected
        with mock.patch('iblutil.io.net.app.asyncio.wait_for', side_effect=asyncio.TimeoutError) as m, \
                self.assertRaises(TimeoutError):
            await self.client.confirmed_send(None, timeout=0.2)
            self.assertFalse(self.client.is_connected, 'failed to close communicator on echo timeout error')
        m.assert_awaited_once_with(mock.ANY, timeout=0.2)
        # Expect to raise RuntimeError with explanation when messages don't match
        with mock.patch('iblutil.io.net.app.asyncio.wait_for', side_effect=RuntimeError), \
                self.assertRaises(RuntimeError) as cm:
            await self.client.confirmed_send(None)
        self.assertIn('unexpected response', str(cm.exception).lower())

    def test_communicator(self):
        """Basic tests for iblutil.io.net.app.EchoProtocol, namely the role setter."""
        # Check role validation
        self.assertEqual(self.server.role, 'server')
        self.assertEqual(self.client.role, 'client')
        with self.assertRaises(ValueError):
            app.EchoProtocol('localhost', 'foo')
        with self.assertRaises(AttributeError):
            self.client.role = 'foo'

    async def test_receive_validation(self):
        """Test for behaviour when non-standard message received."""
        with self.assertWarns(RuntimeWarning), mock.patch.object(self.client, 'send'):
            self.client._receive(b'foo', (self.server.hostname, self.server.port))
        addr = (self.server.hostname, self.server.port)
        fut = asyncio.get_running_loop().create_future()
        self.client._last_sent[addr] = (b'foo', fut)
        with self.assertLogs(self.client.name, logging.ERROR):
            self.client._receive(b'bar', addr)
        self.assertIsInstance(fut.exception(), RuntimeError)
        # Upon receiving message from unknown host, should log warning and return
        with self.assertLogs(self.client.name, logging.WARNING), \
                mock.patch.object(self.client, '_receive') as receive_mock:
            self.client.datagram_received(b'foo', ('192.168.0.0', self.server.port))
            receive_mock.assert_not_called()

    def test_connection_made_validation(self):
        """Test for connection_made method"""
        transport = mock.MagicMock()
        with self.assertRaises(RuntimeError):
            self.client.connection_made(transport)
        transport.get_extra_info().type = socket.SOCK_STREAM
        with self.assertRaises(RuntimeError):
            self.client.connection_made(transport)

    async def test_awaiting_response(self):
        self.assertFalse(self.client.awaiting_response())
        fut = asyncio.get_running_loop().create_future()
        self.client._last_sent[(self.server.hostname, self.server.port)] = (b'foo', fut)
        self.assertTrue(self.client.awaiting_response())
        self.assertFalse(self.client.awaiting_response(addr=('localhost', 8080)))
        fut.cancel()
        self.assertFalse(self.client.awaiting_response())

    async def test_close(self):
        """Test for close/cleanup routine."""
        self.assertTrue(self.client.is_connected)
        loop = asyncio.get_running_loop()
        event_fut = loop.create_future()
        self.client.assign_callback('EXPCLEANUP', event_fut)

        echo_fut = loop.create_future()
        addr = (self.server.hostname, self.server.port)
        self.client._last_sent[addr] = (None, echo_fut)
        self.client.close()

        self.assertFalse(self.client.is_connected)
        self.assertTrue(event_fut.cancelled())
        self.assertTrue(echo_fut.cancelled())
        self.assertFalse(any(self.client._callbacks.values()))
        self.assertTrue(self.client._transport.is_closing())
        self.assertEqual('Close called on communicator', await self.client.on_connection_lost)
        self.assertTrue(self.client.on_eof_received.cancelled())
        self.assertTrue(self.client.on_error_received.cancelled())
        # self.assertEqual(-1, self.client._socket.fileno())  # Closed later on in loop

    async def test_on_error_received(self):
        """Test for on_error_received callback."""
        ex = ValueError('foo')
        with self.assertLogs(self.client.name, logging.ERROR):
            self.client.error_received(ex)
        self.assertTrue(self.client.on_error_received.done())
        self.assertEqual(ex, self.client.on_error_received.result())

    def tearDown(self):
        self.client.close()
        self.server.close()


@unittest.skipIf(ver < Version('3.9'), 'only version 3.9 or later supported')
class TestWebSockets(unittest.IsolatedAsyncioTestCase):
    """Test net.app.EchoProtocol with a TCP/IP transport layer"""

    async def asyncSetUp(self):
        self.server = await app.EchoProtocol.server('ws://localhost:8888', name='server')
        self.client = await app.EchoProtocol.client('ws://localhost:8888', name='client')

    async def test_start(self):
        """Tests confirmed send via start command."""
        # Check socket indeed TCP
        self.assertIs(self.server._socket.type, app.socket.SOCK_STREAM)
        self.assertIs(self.client._socket.type, app.socket.SOCK_STREAM)

        spy = mock.MagicMock()
        self.server.assign_callback('expstart', spy)

        exp_ref = '2022-01-01_1_subject'
        with self.assertLogs(self.server.logger, logging.INFO) as log:
            await self.client.start(exp_ref)
            self.assertIn(f'Received \'[{base.ExpMessage.EXPSTART}, "{exp_ref}", null]', log.records[-1].message)
        spy.assert_called_with([exp_ref, None], (self.client._socket.getsockname()))

    def test_send_validation(self):
        """Test for Communicator.send method."""
        message = b'foo'
        with mock.patch.object(self.client, '_transport') as transport:
            self.client.send(message)
            transport.write.assert_called_with(message)
            transport.write.reset_mock()
            # Check returns when external address used
            with self.assertLogs(self.client.name, logging.WARNING):
                self.client.send(message, addr=('192.168.0.0', 0))
            transport.write.assert_not_called()

    def test_connection_made_validation(self):
        """Test for connection_made method"""
        transport = mock.MagicMock()
        transport.get_extra_info().type = socket.SOCK_DGRAM
        with self.assertRaises(RuntimeError):
            self.client.connection_made(transport)

    def tearDown(self):
        self.client.close()
        self.server.close()


@unittest.skipIf(ver < Version('3.9'), 'only version 3.9 or later supported')
class TestServices(unittest.IsolatedAsyncioTestCase):
    """Tests for the app.Services class"""

    async def asyncSetUp(self):
        # On each acquisition PC
        self.server_1 = await app.EchoProtocol.server('localhost', name='server')
        # On main experiment PC
        self.client_1 = await app.EchoProtocol.client('localhost', name='client1')
        self.client_2 = await app.EchoProtocol.client('localhost', name='client2')
        # For some tests we'll need multiple servers (avoids having to run on multiple threads)
        self.server_2 = await app.EchoProtocol.server('localhost:10002', name='server2')
        self.client_3 = await app.EchoProtocol.client('localhost:10002', name='client3')

    async def test_type(self):
        """Test that services are immutable"""
        services = app.Services([self.client_1, self.client_2])
        # Ensure our services stack is immutable
        with self.assertRaises(TypeError):
            services['client2'] = app.EchoProtocol
        with self.assertRaises(TypeError):
            services.pop('client1')
        # Ensure inputs are validated
        with self.assertRaises(TypeError):
            app.Services([self.client_1, None])

    async def test_close(self):
        """Test Services.close method"""
        clients = [self.client_1, self.client_2]
        assert all(x.is_connected for x in clients)
        services = app.Services(clients)
        self.assertTrue(services.is_connected)
        services.close()
        self.assertFalse(services.is_connected)
        self.assertTrue(not any(x.is_connected for x in clients))

    async def test_assign(self):
        """Tests for Services.assign_callback and Services.clear_callbacks"""
        # Assign a callback for an event
        callback = mock.MagicMock(spec_set=True)
        clients = (self.client_1, self.client_2)
        services = app.Services(clients)
        services.assign_callback('EXPINIT', callback)

        for addr in map(lambda x: x._socket.getsockname(), clients):
            await self.server_1.init('foo', addr=addr)

        self.assertEqual(2, callback.call_count)
        callback.assert_called_with(['foo'], ('127.0.0.1', 11001))

        # Check return_service arg
        callback2 = mock.MagicMock(spec_set=True)
        services.assign_callback('EXPINIT', callback2, return_service=True)
        for addr in map(lambda x: x._socket.getsockname(), clients):
            await self.server_1.init('foo', addr=addr)
        self.assertEqual(2, callback2.call_count)
        callback2.assert_called_with(['foo'], ('127.0.0.1', 11001), self.client_2)

        # Check validation
        with self.assertRaises(TypeError):
            services.assign_callback('EXPEND', 'foo')

        # Check clear callbacks
        services.assign_callback('EXPINIT', callback2)
        removed = services.clear_callbacks('EXPINIT', callback)
        self.assertEqual({'client1': 1, 'client2': 1}, removed)
        removed = services.clear_callbacks('EXPINIT')
        self.assertEqual({'client1': 2, 'client2': 2}, removed)
        # Check futures cancelled
        fut = asyncio.get_running_loop().create_future()
        services.assign_callback('EXPINIT', fut)
        assert not fut.cancelled()
        services.clear_callbacks('EXPINIT')
        self.assertTrue(fut.cancelled())

    async def test_init(self):
        """Test init of services.

        Unfortunately this test is convoluted due to the client and server being on the same
        machine.
        """
        clients = (self.client_1, self.client_3)
        # Require two servers as we'll need two callbacks
        servers = (self.server_1, self.server_2)

        # Set up the client response callbacks that the server (Services object) will await.

        async def respond(server, fut):
            """Response callback for the server"""
            data, addr = await fut
            await asyncio.sleep(.1)  # FIXME Should be able to somehow use loop.call_soon
            await server.init(42, addr)

        for server in servers:
            asyncio.create_task(respond(server, server.on_event(base.ExpMessage.EXPINIT)))

        # Create the services and initialize them, awaiting the callbacks we just set up
        services = app.Services(clients)
        responses = await services.init('foo')

        # Test outcomes
        self.assertFalse(any(map(asyncio.isfuture, responses.values())))
        for name, value in responses.items():
            with self.subTest(client=name):
                self.assertEqual([42], value)

        # Add back the callbacks to test sequential init
        for server in servers:
            asyncio.create_task(respond(server, server.on_event(base.ExpMessage.EXPINIT)))

        # Initialize services sequentially, awaiting the callbacks we just set up
        responses = await services.init('foo', concurrent=False)

        # Test outcomes
        self.assertFalse(any(map(asyncio.isfuture, responses.values())))
        for name, value in responses.items():
            with self.subTest(client=name):
                self.assertEqual([42], value)

    async def test_service_methods(self):
        """Test start, stop, etc. methods.

        For a more complete test, see test_init.
        """
        clients = [mock.AsyncMock(spec=app.EchoProtocol), mock.AsyncMock(spec=app.EchoProtocol)]
        services = app.Services(clients)

        # Init
        await services.init([42, 'foo'])
        for client in clients:
            client.init.assert_awaited_once_with(data=[42, 'foo'])

        # Start
        ref = '2020-01-01_1_subject'
        await services.start(ref)
        for client in clients:
            client.start.assert_awaited_once_with(ref, data=None)

        # Info
        await services.info(base.ExpStatus.RUNNING, {'exp_ref': ref})
        for client in clients:
            client.info.assert_awaited_once_with(base.ExpStatus.RUNNING, data={'exp_ref': ref})

        # Status
        await services.status(base.ExpStatus.STOPPED)
        for client in clients:
            client.status.assert_awaited_once_with(base.ExpStatus.STOPPED)

        # Stop
        await services.stop(immediately=True)
        for client in clients:
            client.stop.assert_awaited_once_with(data=None, immediately=True)

        # Cleanup
        await services.cleanup(data=[42, 'foo'])
        for client in clients:
            client.cleanup.assert_awaited_once_with(data=[42, 'foo'])

        # Alyx
        alyx = mock.MagicMock()
        await services.alyx(alyx)
        for client in clients:
            client.alyx.assert_awaited_once_with(alyx)

    async def test_sequential_signal(self):
        """Test for Services._signal method with concurrent=False"""
        clients = [mock.AsyncMock(spec=app.EchoProtocol), mock.AsyncMock(spec=app.EchoProtocol)]
        for i, client in enumerate(clients):
            client.name = f'client_{i}'
            client.on_event.return_value = ([i], (self.client_1.hostname, self.client_1.port))
        services = app.Services(clients)
        responses = await services._signal(base.ExpMessage.EXPINIT, 'init', 'foo', concurrent=False)
        for client in clients:
            client.init.assert_awaited_once()
        self.assertEqual(responses, {'client_0': [0], 'client_1': [1]})

    def tearDown(self):
        self.client_1.close()
        self.client_2.close()
        self.server_1.close()
        self.server_2.close()
        self.client_3.close()


if __name__ == '__main__':
    from iblutil.util import setup_logger
    setup_logger(app.__name__, level=logging.DEBUG)

    unittest.main()



import sys
from pathlib import Path

from setuptools import setup, find_packages

CURRENT_DIRECTORY = Path(__file__).parent.absolute()

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 8)
VER_ERR_MSG = """
==========================
Unsupported Python version
==========================
This version of iblutil requires Python {}.{}, but you're trying to
install it on Python {}.{}.
"""
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(VER_ERR_MSG.format(*REQUIRED_PYTHON + CURRENT_PYTHON))
    sys.exit(1)

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    require = [x.strip() for x in f.readlines() if not x.startswith('git+')]


def read(rel_path):
    here = Path(__file__).parent.absolute()
    with open(here.joinpath(rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='iblutil',
    version=get_version(Path("iblutil").joinpath("__init__.py")),
    python_requires='>={}.{}'.format(*REQUIRED_PYTHON),
    description='IBL utilities',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='IBL Staff',
    url='https://github.com/int-brain-lab/iblutil/',
    packages=find_packages(exclude=['scratch', 'tests']),  # same as name
    include_package_data=True,
    # external packages as dependencies
    install_requires=require,
    entry_points={},
    scripts=[]
)



import asyncio
import re
import json
import socket
import warnings
import logging
from asyncio import isfuture
from abc import ABC, abstractmethod
from urllib.parse import urlparse
import urllib.request
import ipaddress
from operator import or_
from functools import reduce
from enum import IntFlag, IntEnum, auto  # py3.11 STRICT

LISTEN_PORT = 11001  # listen for commands on this port
PROTOCOL_VERSION = '1.0.0'  # Versioning for ExpMessage, ExpStatus enumerations, and Service base class


def is_success(future: asyncio.Future) -> bool:
    """Check if future successfully resolved."""
    return future.done() and not future.cancelled() and future.exception() is None


def external_ip():
    """
    Fetch WAN IP address.

    NB: Requires internet.

    Returns
    -------
    ipaddress.IPv4Address, ipaddress.IPv6Address
        The computer's default WAN IP address.
    """
    return ipaddress.ip_address(urllib.request.urlopen('https://ident.me').read().decode('utf8'))


def is_valid_ip(ip_address) -> bool:
    """
    Test whether IP address is valid.

    Parameters
    ----------
    ip_address : str
        An IP address to validate.

    Returns
    -------
    bool
        True is IP address is valid.
    """
    try:
        ipaddress.ip_address(ip_address)
        return True
    except ValueError:
        return False


def hostname2ip(hostname=None):
    """
    Resolve hostname to IP address.

    Parameters
    ----------
    hostname : str, optional
        The hostname to resolve.  If None, resolved this computer's hostname.

    Returns
    -------
    ipaddress.IPv4Address, ipaddress.IPv6Address
        The resolved IP address.

    Raises
    ------
    ValueError
        Failed to resolve IP for hostname.
    """
    hostname = hostname or socket.gethostname()
    try:
        ip_address = socket.gethostbyname(hostname)
        return ipaddress.ip_address(ip_address)
    except (socket.error, socket.gaierror):
        raise ValueError(f'Failed to resolve IP for hostname "{hostname}"')


def validate_uri(uri, resolve_host=True, default_port=LISTEN_PORT, default_proc='udp'):
    """
    Ensure URI is complete and correct.

    Parameters
    ----------
    uri : str, ipaddress.IPv4Address, ipaddress.IPv6Address
        A full URI, hostname or hostname and port.
    resolve_host : bool
        If the URI is not an IP address, attempt to resolve hostname to IP.
    default_port : int, str
        If the port is absent from the URI, append this one.
    default_proc : str
        If the URI scheme is missing, prepend this one.

    Returns
    -------
    str
        The complete URI.

    Raises
    ------
    TypeError
        URI type not supported.
    ValueError
        Failed to resolve host name to IP address.
        URI host contains invalid characters (expects only alphanumeric + hyphen).
        Port number not within range (must be > 1, <= 65535).
    """
    # Validate URI scheme
    if not isinstance(uri, (str, ipaddress.IPv4Address, ipaddress.IPv6Address)):
        raise TypeError(f'Unsupported URI "{uri}" of type {type(uri)}')

    if isinstance(uri, str) and (proc := re.match(r'(?P<proc>^[a-zA-Z]+(?=://))', uri)):
        proc = proc.group()
        uri = uri[len(proc) + 3:]
    else:
        proc = default_proc
    # Validate hostname
    if isinstance(uri, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        host = str(uri)
        port = default_port
    elif ':' in uri:
        host, port = uri.split(':', 1)
    else:
        host = uri
        port = None
    if isinstance(uri, str) and not is_valid_ip(host):
        if resolve_host:
            host = hostname2ip(host)
        elif not re.match(r'^[a-z0-9-]+$', host):
            raise ValueError(f'Invalid hostname "{host}"')
    # Validate port
    try:
        port = int(port or default_port)
        assert 1 <= port <= 65535
    except (AssertionError, ValueError):
        raise ValueError(f'Invalid port number: {port or default_port}')
    return f'{proc or default_proc}://{host}:{port}'


# class ExpMessage(IntFlag, boundary=STRICT):  # py3.11
class ExpMessage(IntFlag):
    """A set of standard experiment messages for communicating between rigs."""

    """Experiment is initializing."""
    EXPINIT = auto()
    """Experiment has begun."""
    EXPSTART = auto()
    """Experiment has stopped."""
    EXPEND = auto()
    """Experiment cleanup begun."""
    EXPCLEANUP = auto()
    """Experiment interrupted."""
    EXPINTERRUPT = auto()
    """Experiment status."""
    EXPSTATUS = auto()
    """Experiment info, including task protocol start and end."""
    EXPINFO = auto()
    """Alyx token."""
    ALYX = auto()
    __version__ = PROTOCOL_VERSION

    @classmethod
    def any(cls) -> 'ExpMessage':
        """Return enumeration comprising all possible messages.

        NB: This doesn't include the null ExpMessage (0), used to indicate API errors.
        """
        return reduce(or_, cls)

    @staticmethod
    def validate(event, allow_bitwise=True):
        """
        Validate an event message, returning a corresponding enumeration if valid and raising an
        exception if not.

        Parameters
        ----------
        event : str, int, ExpMessage
            An event message to validate.
        allow_bitwise : bool
            If false, raise if event is the result of a bitwise operation.

        Returns
        -------
        ExpMessage:
            The corresponding event enumeration.

        Raises
        ------
        TypeError
            event is neither a string, integer nor enumeration.
        ValueError
            event does not correspond to any ExpMessage enumeration, neither in its integer form
            nor its string name, or `allow_bitwise` is false and value is combination of events.

        Examples
        --------
        >>> ExpMessage.validate('expstart')
        ExpMessage.EXPSTART

        >>> ExpMessage.validate(10)
        ExpMessage.EXPINIT

        >>> ExpMessage.validate(ExpMessage.EXPEND)
        ExpMessage.EXPEND
        """
        if not isinstance(event, ExpMessage):
            try:
                if isinstance(event, str):
                    event = ExpMessage[event.strip().upper()]
                elif isinstance(event, int):
                    event = ExpMessage(event)
                else:
                    raise TypeError(f'Unknown event type {type(event)}')
            except KeyError:
                raise ValueError(f'Unrecognized event "{event}". '
                                 f'Choices: {tuple(ExpMessage.__members__.keys())}')
        if not allow_bitwise and event not in list(ExpMessage):
            raise ValueError('Compound (bitwise) events not permitted. '
                             f'Choices: {tuple(ExpMessage)}')
        return event

    def __iter__(self):  # py3.11 remove this method
        """Iterate over the individual bits in the enumeration.

        NB: This method is copied from Python 3.11 which supports iteration of Enum objects.
        """
        num = self.value
        while num:
            b = num & (~num + 1)
            yield b
            num ^= b


class ExpStatus(IntEnum):
    """A set of standard statuses for communicating between rigs."""

    """Service is connected."""
    CONNECTED = 0
    """Service is initialized."""
    INITIALIZED = 10
    """Service is running."""
    RUNNING = 20
    """Experiment has stopped."""
    STOPPED = 30
    __version__ = PROTOCOL_VERSION


class Service(ABC):
    """An abstract base class for auxiliary experiment services."""
    __version__ = PROTOCOL_VERSION
    __slots__ = 'name'

    @abstractmethod
    def init(self, data=None):
        """
        Initialize an experiment.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote server.

        Returns
        -------
        ExpMessage.EXPINIT
            The EXPINIT event.
        any, None
            Optional extra data.
        """
        return ExpMessage.EXPINIT, data

    @abstractmethod
    def start(self, exp_ref, data=None):
        """
        Start an experiment.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        exp_ref : str
            An experiment reference string in the form yyyy-mm-dd_n_subject.
        data : any
            Optional extra data to send to the remote server.

        Returns
        -------
        ExpMessage.EXPSTART
            The EXPSTART event.
        str
            The experiment reference string.
        any, None
            Optional extra data.
        """
        exp_ref = exp_ref or None
        if isinstance(exp_ref, dict):
            exp_ref = '_'.join(map(str, (exp_ref['date'], int(exp_ref['sequence']), exp_ref['subject'])))
        return ExpMessage.EXPSTART, exp_ref, data

    @abstractmethod
    def stop(self, data=None, immediately=False):
        """
        Stop an experiment.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote server.
        immediately : bool
            If True, an EXPINTERRUPT message is returned.

        Returns
        -------
        ExpMessage.EXPINTERRUPT, ExpMessage.EXPEND
            The EXPEND event, or EXPINTERRUPT if immediately is True.
        any, None
            Optional extra data.
        """
        return ExpMessage.EXPINTERRUPT if immediately else ExpMessage.EXPEND, data

    @abstractmethod
    def status(self, status):
        """
        Report experiment status.

        NB: This is intended to be lightweight. For more detail and custom data use the info
        method.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        status : ExpStatus
            The experiment status enumeration.

        Returns
        -------
        ExpMessage.EXPSTATUS
            The EXPSTATUS event.
        ExpStatus
            The validated experiment status.
        """
        if not isinstance(status, ExpStatus):
            status = ExpStatus(status) if isinstance(status, int) else ExpStatus[status]
        return ExpMessage.EXPSTATUS, status

    @abstractmethod
    def info(self, status, data=None):
        """
        Report experiment information.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        status : ExpStatus
            The experiment status enumeration.
        data : any
            Optional extra data to send to the remote server.

        Returns
        -------
        ExpMessage.EXPINFO
            The EXPINFO event.
        ExpStatus
            The validated experiment status.
        any, None
            Optional extra data.
        """
        return ExpMessage.EXPINFO, Service.status(self, status)[1], data

    @abstractmethod
    def cleanup(self, data=None):
        """
        Clean up an experiment.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        data : any
            Optional extra data to send to the remote server.

        Returns
        -------
        ExpMessage.EXPCLEANUP
            The EXPCLEANUP event.
        any, None
            Optional extra data.
        """
        return ExpMessage.EXPCLEANUP, data

    @abstractmethod
    def alyx(self, alyx):
        """
        Request/send Alyx token.

        This is intended to specify the expected message signature. The subclassed method should
        serialize the returned values and pass them to the transport layer.

        Parameters
        ----------
        alyx : one.webclient.AlyxClient
            Optional instance of Alyx to send.

        Returns
        -------
        ExpMessage.ALYX
            The ALYX event.
        str
            The Alyx database URL.
        dict
            The Alyx token in the form {user: token}.
        """
        base_url = alyx.base_url if alyx else None
        token = {alyx.user: alyx._token} if alyx and alyx.is_logged_in else {}
        return ExpMessage.ALYX, base_url, token


class Communicator(Service):
    """A base class for communicating between experimental rigs.

    Attributes
    ----------
    name : str
        An arbitrary label for the remote host
    server_uri : str
        The full URI of the remote device, e.g. udp://192.168.0.1:1001
    """
    __slots__ = ('server_uri', '_callbacks', 'logger', 'name')

    def __init__(self, server_uri, name=None, logger=None):
        self.server_uri = validate_uri(server_uri)
        self.name = name or server_uri
        self.logger = logger or logging.getLogger(self.name)
        # Init callbacks map of ExpMessage -> list, including null ExpMessage for processing callback errors
        self._callbacks = dict(map(lambda item: (item, []), (ExpMessage(0), *ExpMessage.__members__.values())))

    def assign_callback(self, event, callback):
        """
        Assign a callback to be called when an event occurs.

        NB: Unlike with futures, an assigned callback may be triggered multiple times, whereas
        coroutines may only be set once after which they are cleared.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            The event for which the callback is registered.
        callback : function, asyncio.Future
            A function or Future to trigger when an event occurs.

        See Also
        --------
        EchoProtocol.receive
            The method that processes the callbacks upon receiving a message.
        """
        event = ExpMessage.validate(event)
        if not (callable(callback) or isfuture(callback)):
            raise TypeError('Callback must be callable or a Future')
        if event is ExpMessage(0):
            self._callbacks.setdefault(event, []).append((callback, False))
        else:
            return_event = event not in ExpMessage
            for e in event:  # iterate over enum as bitwise ops permitted
                self._callbacks.setdefault(e, []).append((callback, return_event))

    def clear_callbacks(self, event, callback=None, cancel_futures=True):
        """
        For a given event, remove the provided callback, or all callbacks if none were provided.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            The event for which the callback was registered.
        callback : function, asyncio.Future
            The callback or future to remove.
        cancel_futures : bool
            If True and callback is a Future, cancel before removing.

        Returns
        -------
        int
            The number of callbacks removed.
        """
        i = 0
        event = ExpMessage.validate(event)
        if callback:  # clear specific callback
            # Wrapped callables have an id attribute containing the hash of the inner function
            for evt in event:  # iterate as bitwise enums permitted, e.g. ~ExpMessage.ALYX
                ids = [getattr(cb, 'id', None) or hash(cb) for cb, _ in self._callbacks[evt]]
                cb_id = getattr(callback, 'id', None) or hash(callback)
                while True:
                    try:
                        idx = ids.index(cb_id)
                        if cancel_futures and isfuture(cb := self._callbacks[evt][idx][0]):
                            cb.cancel()
                        del self._callbacks[evt][idx]
                        del ids[idx]
                        i += 1
                    except (IndexError, ValueError):
                        break
        else:  # clear all callbacks for event
            for evt in event:
                if cancel_futures:
                    for cb in filter(isfuture, map(lambda x: x[0], self._callbacks[evt])):
                        cb.cancel()
                i += len(self._callbacks[evt])
                del self._callbacks[evt][:]
        self.logger.debug('[%s] %i callbacks cleared', self.name, i)
        return i

    async def on_event(self, event):
        """
        Await an event from the remote host.

        Parameters
        ----------
        event : str, int, iblutil.io.net.base.ExpMessage
            The event to wait on.

        Returns
        -------
        any
            The response data returned by the remote host.

        Examples
        --------
        >>> data, addr = await com.on_event('EXPSTART')

        >>> event = await asyncio.create_task(com.on_event('EXPSTART'))
        >>> ...
        >>> data = await event

        Await more than one event

        >>> data, addr, event = await com.on_event(ExpMessage.EXPEND | ExpMessage.EXPINTERRUPT)
        """
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        self.assign_callback(event, fut)
        return await fut

    @property
    def port(self) -> int:
        """int: the remote port"""
        return int(urlparse(self.server_uri).port)

    @property
    def hostname(self) -> str:
        """str: the remote hostname or IP address"""
        return urlparse(self.server_uri).hostname

    @property
    def protocol(self) -> str:
        """str: the protocol scheme, e.g. udp, ws"""
        return urlparse(self.server_uri).scheme

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """bool: True if the remote device is connected"""
        pass

    @abstractmethod
    def send(self, data, addr=None):
        """Serialize and pass data to the transport layer"""
        pass

    def _receive(self, data, addr):
        """
        Process data received from remote host and notify event listeners.

        This is called by the transport layer when a message is received and should not be called
        by the user.

        Parameters
        ----------
        data : bytes
            The serialized data received by the transport layer.
        addr : (str, int)
            The source address as (hostname, port)

        Warnings
        --------
        Warnings
            Expects the deserialized data to be a tuple where the first element is an ExpMessage.
            A warning is thrown if the data is not a tuple, or has fewer than two elements.

        TODO Perhaps for every event only the first future should be set.
        """
        data = self.decode(data)
        if isinstance(data, (list, tuple)) and len(data) > 1:
            event, *data = data
            event = ExpMessage.validate(event, allow_bitwise=False) if event else ExpMessage(0)
            if event is ExpMessage(0):
                # An error in the remote callback function occurred
                err, evt = data
                self.logger.error('Callback for %s on %s://%s:%i failed with %s',
                                  ExpMessage(evt).name, self.protocol, *addr, err)
            for f, return_event in self._callbacks[event].copy():
                if isfuture(f):
                    if f.done():
                        self.logger.warning('Future %s already resolved', f)
                    elif not f.cancelled():
                        f.set_result((data, addr, event) if return_event else (data, addr))
                    self.clear_callbacks(event, f)  # Remove future from list
                else:
                    try:
                        f(data, addr, event) if return_event else f(data, addr)
                    except Exception as ex:
                        self.logger.error('Callback "%s" failed with error "%s"', f, ex)
                        message = self.encode([0, f'{type(ex).__name__}: {ex}', event])
                        self.send(message, addr)
                        break
        else:
            warnings.warn(f'Expected list, got {data}', RuntimeWarning)

    @staticmethod
    def encode(data) -> bytes:
        """
        Serialize data for transmission.

        None-string or -bytes objects are encoded as JSON before converting to bytes.

        Parameters
        ----------
        data : any
            The data to serialize.

        Returns
        -------
        bytes
            The encoded data.
        """
        if isinstance(data, bytes):
            return data
        if not isinstance(data, str):
            data = json.dumps(data)
        return data.encode()

    @staticmethod
    def decode(data: bytes):
        """
        De-serialize and parse bytes data.

        This function attempts to decode the data as a JSON object.  On failing that, returns a
        string.

        Parameters
        ----------
        data : bytes
            The data to de-serialize.

        Returns
        -------
        any
            Deserialized data.
        """
        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            warnings.warn('Failed to decode as JSON')
            data = data.decode()
        return data

    def close(self):
        """De-register all callbacks and cancel futures"""
        for event in self._callbacks:
            for fut in filter(isfuture, map(lambda x: x[0], self._callbacks[event])):
                fut.cancel('Close called on communicator')
            del self._callbacks[event][:]



import hashlib
from typing import TypeVar, Sequence, Union, Optional, Type
import uuid

import numpy as np

D = TypeVar("D", bound=np.generic)
Array = Union[np.ndarray, Sequence]


def between_sorted(sorted_v, bounds=None):
    """
    Given a vector of sorted values, returns a boolean vector indices True when the value
    is between bounds. If multiple bounds are given, returns the equivalent OR of individual
    bounds tuple
    Especially useful for spike times
      indices = between_sorted(spike_times, [tstart, tstop])
    :param sorted_v: vector containing sorted values (won't check)
    :param bounds: minimum included value and maximum included value
        can be a list[tstart, tstop] or an array of dimension (n, 2)
    :return:
    """
    bounds = np.array(bounds)
    starts, stops = (np.take(bounds, 0, axis=-1), np.take(bounds, 1, axis=-1))
    sbounds = np.logical_and(starts <= sorted_v[-1], stops >= sorted_v[0])
    starts = starts[sbounds]
    stops = stops[sbounds]
    sel = sorted_v * 0
    sel[np.searchsorted(sorted_v, starts)] = 1
    istops = np.searchsorted(sorted_v, stops, side="right")
    sel[istops[istops < sorted_v.size]] += -1
    return np.cumsum(sel).astype(bool)


def hash_uuids(uuids, algo="sha256"):
    """
    Returns a hash of an array or list of UUID strings
    :param uuids: collection of UUID strings: list | np.ndarray | pd.Series | pd.Index
    :param algo: hashing algorithm to use, default is sha256
    """
    hash = hashlib.new(algo)
    for cid in uuids:
        _uuid = cid if isinstance(cid, uuid.UUID) else uuid.UUID(hex=cid)
        hash.update(_uuid.bytes)
    return hash.hexdigest()


def ismember(a, b):
    """
    equivalent of np.isin but returns indices as in the matlab ismember function
    returns an array containing logical 1 (true) where the data in A is B
    also returns the location of members in b such as a[lia] == b[locb]
    :param a: 1d - array
    :param b: 1d - array
    :return: isin, locb
    """
    lia = np.isin(a, b)
    aun, _, iuainv = np.unique(a[lia], return_index=True, return_inverse=True)
    _, ibu, iau = np.intersect1d(b, aun, return_indices=True)
    locb = ibu[iuainv]
    return lia, locb


def ismember2d(a, b):
    """
    Equivalent of np.isin but returns indices as in the matlab ismember function
    returns an array containing logical 1 (true) where the data in A is B
    also returns the location of members in b such as a[lia, :] == b[locb, :]
    :param a: 2d array
    :param b: 2d array
    :return: isin, locb
    """
    from numba import jit

    @jit(nopython=True)
    def find_first_2d(mat, val):
        """
        Returns first index where
        The purpose of this function is performance: uses low level numba and avoids looping
        through the full array
        :param mat: np.array
        :param val: values to search for
        :return: index or empty array
        """
        for i in np.arange(mat.shape[0]):
            if np.all(mat[i] == val):
                return i

    amask = np.ones(a.shape[0], dtype=bool)
    ia = np.zeros(a.shape, dtype=bool)
    ib = np.zeros(a.shape, dtype=np.int32) - 1
    ina = np.zeros(a.shape[0], dtype=bool)
    bind = np.zeros(a.shape[0], dtype=np.int32) - 1
    # get a 1d ismember first for each column
    for n in np.arange(a.shape[1]):
        iaa, ibb = ismember(a[amask, n], b[:, n])
        ia[amask, n] = iaa
        ib[np.where(amask)[0][iaa], n] = ibb
    # those that have at least one mismatch are not in
    amask[~np.all(ia, axis=1)] = False
    # those that get the same index for all column do not need further testing
    ifound = np.where(amask)[0][np.sum(np.abs(np.diff(ib[amask], axis=1)), axis=1) == 0]
    ina[ifound] = True
    amask[ifound] = False
    bind[ifound] = ib[ifound, 0]
    # the remaining ones have to be check manually (almost never happens for uuids)
    for iaa in np.where(amask)[0]:
        ibb = find_first_2d(b, a[iaa, :])
        if ibb:
            ina[iaa] = True
            bind[iaa] = ibb
    return ina, bind[ina]


def intersect2d(a0, a1, assume_unique=False):
    """
    Performs intersection on multiple columns arrays a0 and a1
    :param a0:
    :param a1:
    :param assume_unique: If True, the input arrays are both assumed to be unique,
    which can speed up the calculation.
    :return: intersection
    :return: index of a0 such as intersection = a0[ia, :]
    :return: index of b0 such as intersection = b0[ib, :]
    """
    _, i0, i1 = np.intersect1d(
        a0[:, 0], a1[:, 0], return_indices=True, assume_unique=assume_unique
    )
    for n in np.arange(1, a0.shape[1]):
        _, ii0, ii1 = np.intersect1d(
            a0[i0, n], a1[i1, n], return_indices=True, assume_unique=assume_unique
        )
        i0 = i0[ii0]
        i1 = i1[ii1]
    return a0[i0, :], i0, i1


def bincount2D(x, y, xbin=0, ybin=0, xlim=None, ylim=None, weights=None):
    """
    Computes a 2D histogram by aggregating values in a 2D array.

    :param x: values to bin along the 2nd dimension (c-contiguous)
    :param y: values to bin along the 1st dimension
    :param xbin:
        scalar: bin size along 2nd dimension
        0: aggregate according to unique values
        array: aggregate according to exact values (count reduce operation)
    :param ybin:
        scalar: bin size along 1st dimension
        0: aggregate according to unique values
        array: aggregate according to exact values (count reduce operation)
    :param xlim: (optional) 2 values (array or list) that restrict range along 2nd dimension
    :param ylim: (optional) 2 values (array or list) that restrict range along 1st dimension
    :param weights: (optional) defaults to None, weights to apply to each value for aggregation
    :return: 3 numpy arrays MAP [ny,nx] image, xscale [nx], yscale [ny]
    """
    # if no bounds provided, use min/max of vectors
    if xlim is None:
        xlim = [np.min(x), np.max(x)]
    if ylim is None:
        ylim = [np.min(y), np.max(y)]

    def _get_scale_and_indices(v, bin, lim):
        # if bin is a nonzero scalar, this is a bin size: create scale and indices
        if np.isscalar(bin) and bin != 0:
            scale = np.arange(lim[0], lim[1] + bin / 2, bin)
            ind = (np.floor((v - lim[0]) / bin)).astype(np.int64)
        # if bin == 0, aggregate over unique values
        else:
            scale, ind = np.unique(v, return_inverse=True)
        return scale, ind

    xscale, xind = _get_scale_and_indices(x, xbin, xlim)
    yscale, yind = _get_scale_and_indices(y, ybin, ylim)
    # aggregate by using bincount on absolute indices for a 2d array
    nx, ny = [xscale.size, yscale.size]
    ind2d = np.ravel_multi_index(np.c_[yind, xind].transpose(), dims=(ny, nx))
    r = np.bincount(ind2d, minlength=nx * ny, weights=weights).reshape(ny, nx)

    # if a set of specific values is requested output an array matching the scale dimensions
    if not np.isscalar(xbin) and xbin.size > 1:
        _, iout, ir = np.intersect1d(xbin, xscale, return_indices=True)
        _r = r.copy()
        r = np.zeros((ny, xbin.size))
        r[:, iout] = _r[:, ir]
        xscale = xbin

    if not np.isscalar(ybin) and ybin.size > 1:
        _, iout, ir = np.intersect1d(ybin, yscale, return_indices=True)
        _r = r.copy()
        r = np.zeros((ybin.size, r.shape[1]))
        r[iout, :] = _r[ir, :]
        yscale = ybin

    return r, xscale, yscale


def rcoeff(x, y):
    """
    Computes pairwise Pearson correlation coefficients for matrices.

    That is for 2 matrices the same size, computes the row to row coefficients and outputs
    a vector corresponding to the number of rows of the first matrix.
    If the second array is a vector then computes the correlation coefficient for all rows.
    :param x: np array [nc, ns]
    :param y: np array [nc, ns] or [ns]
    :return: r [nc]
    """

    def normalize(z):
        mean = np.mean(z, axis=-1)
        return z - mean if mean.size == 1 else z - mean[:, np.newaxis]

    xnorm = normalize(x)
    ynorm = normalize(y)
    rcor = np.sum(xnorm * ynorm, axis=-1) / np.sqrt(
        np.sum(np.square(xnorm), axis=-1) * np.sum(np.square(ynorm), axis=-1)
    )
    return rcor


def within_ranges(
    x: np.ndarray,
    ranges: Array,
    labels: Optional[Array] = None,
    mode: str = "vector",
    dtype: Type[D] = "int8",
) -> np.ndarray:
    """
    Detects which points of the input vector lie within one of the ranges specified in the ranges.
    Returns an array the size of x with a 1 if the corresponding point is within a range.

    The function uses a stable sort algorithm (timsort) to find the edges within the input array.
    Edge behaviour is inclusive.

    Ranges are [(start0, stop0), (start1, stop1), etc.] or n-by-2 numpy array.
    The ranges may be optionally assigned a row in 'matrix' mode or a numerical label in 'vector'
    mode. Labels must have a length of n.  Overlapping ranges have a value that is the sum of the
    relevant range labels (ones in 'matrix' mode).

    If mode is 'vector' (default) it will give a vector, specifying the range of each point.
    If mode is 'matrix' it will give a matrix output where each range is assigned a particular row
    index with 1 if the point belongs to that range label.  Multiple ranges can be assigned to a
    particular row, e.g. [0, 0,1] would give a 2-by-N matrix with the first two ranges in the
    first row.  Points within more than one range are given a value > 1

    Parameters
    ----------
    x : array_like
        An array whose points are tested against the ranges.  multi-dimensional arrays are
        flattened to 1D
    ranges : array_like
        A list of tuples or N-by-2 array of ranges to test, where N is the number of ranges,
        i.e. [[start0, stop0],
        [start1, stop1]]
    labels : vector, list
        If mode is 'vector'; a list of integer labels to demarcate which points lie within each
        range.  In 'matrix' mode; a list of column indices (ranges can share indices).
        The number of labels should match the number of ranges.  If None, ones are used for all
        ranges.
    mode : {'matrix', 'vector'}
        The type of output to return.  If 'matrix' (default), an N-by-M matrix is returned where N
        is the size of x and M corresponds to the max index in labels, e.g. with labels=[0,1,2],
        the output matrix would have 3 columns.  If 'vector' a vector the size of x is returned.
    dtype : str, numeric or boolean type
        The data type of the returned array.  If type is bool, the labels in vector mode will be
        ignored.  Default is int8.


    Returns
    -------
    A vector of size like x where zeros indicate that the points do not lie within ranges (
    'vector' mode) or a matrix where out.shape[0] == x.size and out.shape[1] == max(labels) + 1.

    Examples
    -------
    # Assert that points in ranges are mutually exclusive
    np.all(within_ranges(x, ranges) <= 1)

    Tests
    -----
    >>> import numpy as np
    >>> within_ranges(np.arange(11), [(1, 2), (5, 8)])
    array([0, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0], dtype=int8)
    >>> ranges = np.array([[1, 2], [5, 8]])
    >>> within_ranges(np.arange(10) + 1, ranges, labels=np.array([0,1]), mode='matrix')
    array([[1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 1, 1, 1, 0, 0]], dtype=int8)
    >>> within_ranges(np.arange(11), [(1,2), (5,8), (4,6)], labels=[0,1,1], mode='matrix')
    array([[0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 1, 2, 2, 1, 1, 0, 0]], dtype=int8)
    >>> within_ranges(np.arange(10) + 1, ranges, np.array([3,1]), mode='vector')
    array([3, 3, 0, 0, 1, 1, 1, 1, 0, 0], dtype=int8)
    >>> within_ranges(np.arange(11), [(1,2), (5,8), (4,6)], dtype=bool)
    array([False,  True,  True, False,  True,  True,  True,  True,  True,
           False, False])
    """
    # Flatten
    x = x.ravel()

    # Ensure ranges are numpy
    ranges = np.array(ranges)

    # Get size info
    n_points = x.size
    n_ranges = ranges.shape[0]

    if labels is None:
        # In 'matrix' mode default row index is 0
        labels = np.zeros((n_ranges,), dtype="uint32")
        if mode == "vector":  # Otherwise default numerical label is 1
            labels += 1
    assert len(labels) >= n_ranges, "range labels do not match number of ranges"
    n_labels = np.unique(labels).size

    # If no ranges given, short circuit function and return zeros
    if n_ranges == 0:
        return np.zeros_like(x, dtype=dtype)

    # Check end comes after start in each case
    assert np.all(
        np.diff(ranges, axis=1) > 0
    ), "ranges ends must all be greater than starts"

    # Make array containing points, starts and finishes

    # This order means it will be inclusive
    to_sort = np.concatenate((ranges[:, 0], x, ranges[:, 1]))
    # worst case O(n*log(n)) but will be better than this as most of the array is ordered;
    # memory overhead ~n/2
    idx = np.argsort(to_sort, kind="stable")

    # Make delta array containing 1 for every start and -1 for every stop
    # with one row for each range label
    if mode == "matrix":
        delta_shape = (n_labels, n_points + 2 * n_ranges)
        delta = np.zeros(delta_shape, dtype="int8")

        delta[labels, np.arange(n_ranges)] = 1
        delta[labels, n_points + n_ranges + np.arange(n_ranges)] = -1

        # Arrange in order
        delta_sorted = delta[:, idx]

        # Take cumulative sums
        summed = np.cumsum(delta_sorted, axis=1)

        # Reorder back to original order
        reordered = np.zeros(delta_shape, dtype=dtype)
        reordered[:, idx] = summed.reshape(delta_shape[0], -1)
        return reordered[:, np.arange(n_ranges, n_points + n_ranges)]

    elif mode == "vector":
        delta_shape = (n_points + 2 * n_ranges,)
        r_delta = np.zeros(delta_shape, dtype="int32")
        r_delta[np.arange(n_ranges)] = labels
        r_delta[n_points + n_ranges + np.arange(n_ranges)] = -labels

        # Arrange in order
        r_delta_sorted = r_delta[idx]

        # Take cumulative sum
        r_summed = np.cumsum(r_delta_sorted)

        # Reorder back to original
        r_reordered = np.zeros_like(r_summed, dtype=dtype)
        r_reordered[idx] = r_summed

        return r_reordered[np.arange(n_ranges, n_points + n_ranges)]

    else:
        raise ValueError('unknown mode type, options are "matrix" and "vector"')



"""Tools to generate and identify spacers.

Spacers are sequences of up and down pulses with a specific, identifiable pattern.
They are generated with a chirp coding to reduce cross-correlaation sidelobes.
They are used to mark the beginning of a behaviour sequence within a session.

Example
-------
>>> spacer = Spacer()
>>> spacer.add_spacer_states(sma, t, next_state='first_state')
>>> for i in range(ntrials):
... sma.add_state(
...     state_name='first_state',
...     state_timer=tup,
...     state_change_conditions={'Tup': f'spacer_low_{i:02d}'},
...     output_actions=[('BNC1', 255)],  # To FPGA
... )

"""

import numpy as np


class Spacer:
    def __init__(self, dt_start=.02, dt_end=.4, n_pulses=8, tup=.05):
        """Computes spacer up times using a chirp up and down pattern.

        Parameters
        ----------
        dt_start : float
            First spacer up time.
        dt_end : float
            Last spacer up time.
        n_pulses : int
            Number of spacer up times, one-sided (i.e. 8 means 16 - 1 spacers times)
        tup: float
            Duration of the spacer up time.
        """
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.n_pulses = n_pulses
        self.tup = tup
        assert np.all(np.diff(self.times) > self.tup), 'Spacers are overlapping'

    def __repr__(self):
        return f'Spacer(dt_start={self.dt_start}, dt_end={self.dt_end}, n_pulses={self.n_pulses}, tup={self.tup})'

    @property
    def times(self):
        """Computes spacer up times using a chirp up and down pattern.

        Each time corresponds to an up time of the BNC1 signal.

        Returns
        -------
        numpy.array
            Numpy arrays of spacer times.
        """
        # upsweep
        t = np.linspace(self.dt_start, self.dt_end, self.n_pulses) + self.tup
        # downsweep
        t = np.r_[t, np.flipud(t[1:])]
        t = np.cumsum(t)
        return t

    def generate_template(self, fs=1000):
        """
        Generates a spacer voltage template to cross-correlate with a voltage trace from a DAQ to
        detect a voltage trace.

        Parameters
        ----------
        fs : int
            DAQ sampling frequency.

        Returns
        -------
        numpy.array
            The template spacer signal.
        """
        t = self.times
        ns = int((t[-1] + self.tup * 10) * fs)
        sig = np.zeros(ns, )
        sig[(t * fs).astype(np.int32)] = 1
        sig[((t + self.tup) * fs).astype(np.int32)] = -1
        sig = np.cumsum(sig)
        return sig

    def add_spacer_states(self, sma=None, next_state='exit'):
        """
        Add spacer states to a state machine.

        Parameters
        ----------
        sma : pybpodapi.state_machine.StateMachine
            A Bpod state machine instance.
        next_state : str
            The name of the state to follow the spacer state.
        """
        assert next_state is not None
        t = self.times
        dt = np.diff(t, append=t[-1] + self.tup * 2)
        for i, time in enumerate(t):
            if sma is None:
                print(i, time, dt[i])
                continue
            next_loop = f'spacer_high_{i + 1:02d}' if i < len(t) - 1 else next_state
            sma.add_state(
                state_name=f'spacer_high_{i:02d}',
                state_timer=self.tup,
                state_change_conditions={'Tup': f'spacer_low_{i:02d}'},
                output_actions=[('BNC1', 255)],  # To FPGA
            )
            sma.add_state(
                state_name=f'spacer_low_{i:02d}',
                state_timer=dt[i] - self.tup,
                state_change_conditions={'Tup': next_loop},
                output_actions=[],
            )

    def find_spacers_from_fronts(self, fronts, fs=1000):
        """
        Given the timestamps and polarities of a digital signal, returns the timestamps of each
        signal.  This method first finds the locations where there are n consecutive pulses of the
        correct width then convolves this part of the signal with the template signal.

        This method may be relaxed in order to make it robust to noise in the signal.

        Parameters
        ----------
        fronts : dict[str, numpy.array]
            Dictionary with keys ('times', 'polarities') containing the timestamps and polarities
            of the signal fronts, respectively.
        fs : int
            The sampling frequency of the DAQ signal.

        Returns
        -------
        numpy.array
            The times of the protocol spacer signals.
        """
        n_pulses = (self.n_pulses * 2) - 1
        is_pulse = np.isclose(np.diff(fronts['times']), self.tup, rtol=1e-2)
        is_pulse = np.insert(is_pulse, 0, False)
        ind, = np.where(is_pulse)

        # Find consecutive pulses that are the correct length close together
        max_d = 1.  # look for fronts less than 1 second apart
        consecutive = np.logical_and(np.diff(ind) == 2, np.diff(fronts['times'][ind]) < max_d)
        consecutive = np.pad(consecutive, 1, 'constant', constant_values=False)
        edges, = np.where(~consecutive)
        spacer_times = []
        for i in np.arange(edges.size - 1):
            if edges[i + 1] - edges[i] == n_pulses:  # This could be relaxed to allow for noise
                idx = np.arange(ind[edges[i]], ind[edges[i + 1] - 1] + 1)  # +1 to include final down
                t = fronts['times'][idx]
                ts = np.arange(t[0], t[-1], 1 / fs)  # Evenly resample at given frequency
                # Reconstruct trace where 1 = high, 0 = low
                signal = np.zeros_like(ts)
                ii = np.searchsorted(ts, t, side='left')
                signal[ii[ii < len(signal)]] = fronts['polarities'][idx[ii < len(signal)]]
                signal = np.cumsum(signal) + 1  # {-1, 0} -> {0, 1}
                try:
                    spacer, = self.find_spacers(signal, fs=fs)
                    spacer_times.append(spacer + t[0])
                except IndexError:
                    continue

        return np.array(spacer_times)

    def find_spacers(self, signal, threshold=0.9, fs=1000):
        """
        Find spacers in a voltage time series. Assumes that the signal is a digital signal between
        0 and 1.

        Parameters
        ----------
        signal : numpy.ndarray
            The signal in which to find the spacer.
        threshold : float
            The cross-correlation detection threshold.
        fs : int
            The sampling frequency of the DAQ signal.

        Returns
        -------
        numpy.ndarray
            An array containing the times of each spacer signal relative to the first sample.
        """
        template = self.generate_template(fs=fs)
        xcor = np.correlate(signal, template, mode='full') / np.sum(template)
        idetect = np.where(xcor > threshold)[0]
        iidetect = np.cumsum(np.diff(idetect, prepend=0) > 1)
        nspacers = iidetect[-1]
        tspacer = np.zeros(nspacers)
        for i in range(nspacers):
            ispacer = idetect[iidetect == i + 1]
            imax = np.argmax(xcor[ispacer])
            tspacer[i] = (ispacer[imax] - template.size + 1) / fs
        return tspacer
